from compmake.ui.helpers import ui_command, VISUALIZATION
from compmake.jobs.storage import job_cache_exists, get_job_cache
from compmake.structures import Cache
from compmake.utils.colored import compmake_colored


@ui_command(section=VISUALIZATION)
def why(non_empty_job_list, context, cq):
    """ Shows the last line of the error """
    lines = []
    for job_id in non_empty_job_list:
        details = details_why_one(job_id, context, cq)

        if details is not None:
            lines.append(details)

    s = format_table(lines)
    print(s)


def format_table(lines, sep=" | "):
    """ lines is a list of tuples """
    if not lines:
        return ""
    ncols = len(lines[0])
    cols = [list(_[i] for _ in lines) for i in range(ncols)]
    maxchars = lambda col: max(len(_) for _ in col)
    maxc = list(map(maxchars, cols))

    s = ""
    for line in lines:
        for i in range(ncols):
            spec = "%%-%ds" % maxc[i]
            cell = spec % line[i]
            if "NotImplementedError" in cell:
                cell = compmake_colored(cell, color="blue", attrs=[])
            s += cell
            if i < ncols - 1:
                s += sep
        s += "\n"
    return s


def details_why_one(job_id, context, cq):
    db = context.get_compmake_db()

    lines = []
    if job_cache_exists(job_id, db):
        cache = get_job_cache(job_id, db)

        status = Cache.state2desc[cache.state]
        if cache.state in [Cache.FAILED, Cache.BLOCKED]:
            why = cache.exception
            why = why.strip()
            lines = why.split("\n")
            one = lines[0]
            if len(lines) > 1:
                one += " [+%d lines] " % (len(lines) - 1)

            details = (job_id, status, one)
            return details

    return None
    # print('%s20s: %s' %(job_id, one))



import sys
import traceback


def wrap_script_entry_point(function, exceptions_no_traceback):
    """
        Wraps the main() of a script.
        For Exception: we exit with value 2.

        :param exceptions_no_traceback: tuple of exceptions for which we
         just print the error, and return 1.

    """
    try:
        ret = function(sys.argv[1:])
        if ret is None:
            ret = 0
        sys.exit(ret)
    except exceptions_no_traceback as e:
        print(str(e))
        sys.exit(1)
    except Exception as e:
        print(traceback.format_exc())
        sys.exit(2)



import signal
import traceback
from multiprocessing import TimeoutError
from multiprocessing.context import BaseContext

from future.moves.queue import Empty

from compmake.exceptions import CompmakeBug, HostFailed, JobFailed, JobInterrupted
from compmake.jobs.manager import AsyncResultInterface
from compmake.jobs.result_dict import result_dict_raise_if_error
from zuper_commons.text import indent
from zuper_commons.types import check_isinstance

__all__ = [
    "PmakeSub",
]


class PmakeSub:
    EXIT_TOKEN = "please-exit"

    def __init__(self, name: str, signal_queue, signal_token, ctx: BaseContext, write_log=None):
        self.name = name

        self.job_queue = ctx.Queue()
        self.result_queue = ctx.Queue()
        # print('starting process %s ' % name)
        self.proc = ctx.Process(
            target=pmake_worker,
            args=(self.name, self.job_queue, self.result_queue, signal_queue, signal_token, write_log),
            name=name,
        )
        self.proc.start()

    def terminate(self):
        self.job_queue.put(PmakeSub.EXIT_TOKEN)
        self.job_queue.close()
        self.result_queue.close()
        self.job_queue = None
        self.result_queue = None

    def apply_async(self, function, arguments):
        self.job_queue.put((function, arguments))
        self.last = PmakeResult(self.result_queue)
        return self.last


def pmake_worker(name, job_queue, result_queue, signal_queue, signal_token, write_log=None):
    # logger.info(f"pmake_worker forked at process {os.getpid()}")
    from coverage import process_startup

    if hasattr(process_startup, "coverage"):
        # logger.info("Detected coverage wanted.")
        delattr(process_startup, "coverage")
        cov = process_startup()
        if cov is None:
            pass
            # logger.warning("Coverage did not start.")
        else:
            pass
            # logger.info("Coverage started successfully.")
    else:
        # logger.info("Not detected coverage need.")
        cov = None

    if write_log:
        f = open(write_log, "w")

        def log(s):
            f.write(f"{name}: {s}\n")
            f.flush()

    else:

        def log(s):
            print(f"{name}: {s}")
            pass

    log("started pmake_worker()")
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    def put_result(x):
        log("putting result in result_queue..")
        result_queue.put(x, block=True)
        if signal_queue is not None:
            log("putting result in signal_queue..")
            signal_queue.put(signal_token, block=True)
        log("(done)")

    try:
        while True:
            log("Listening for job")
            try:
                job = job_queue.get(block=True, timeout=5)
            except Empty:
                log("Could not receive anything.")
                continue
            if job == PmakeSub.EXIT_TOKEN:
                log("Received EXIT_TOKEN.")
                break

            log(f"got job: {job}")

            function, arguments = job
            try:
                # print('arguments: %s' % str(arguments))
                result = function(arguments)
            except JobFailed as e:
                log("Job failed, putting notice.")
                log(f"result: {e}")  # debug
                put_result(e.get_result_dict())
            except JobInterrupted as e:
                log("Job interrupted, putting notice.")
                put_result(dict(abort=str(e)))  # XXX
            except CompmakeBug as e:  # XXX :to finish
                log("CompmakeBug")
                put_result(e.get_result_dict())
            else:
                log(f"result: {result}")
                put_result(result)

            log("...done.")

            # except KeyboardInterrupt: pass
    except BaseException as e:
        reason = "aborted because of uncaptured:\n" + indent(traceback.format_exc(), "| ")
        mye = HostFailed(host="???", job_id="???", reason=reason, bt=traceback.format_exc())
        log(str(mye))
        put_result(mye.get_result_dict())
    except:
        mye = HostFailed(
            host="???", job_id="???", reason="Uknown exception (not BaseException)", bt="not available"
        )
        log(str(mye))
        put_result(mye.get_result_dict())
        log("(put)")

    if signal_queue is not None:
        signal_queue.close()
    result_queue.close()
    log("saving coverage")
    if cov:
        cov._atexit()
    log("saved coverage")

    log("clean exit.")


class PmakeResult(AsyncResultInterface):
    """ Wrapper for the async result object obtained by pool.apply_async """

    def __init__(self, result_queue):
        self.result_queue = result_queue
        self.result = None
        # self.count = 0

    def ready(self):
        # self.count += 1
        try:
            self.result = self.result_queue.get(block=False)
        except Empty:
            # if self.count > 1000 and self.count % 100 == 0:
            # print('ready()?  still waiting on %s' % str(self.job))
            return False
        else:
            return True

    def get(self, timeout=0):
        if self.result is None:
            try:
                self.result = self.result_queue.get(block=True, timeout=timeout)
            except Empty as e:
                raise TimeoutError(e)

        check_isinstance(self.result, dict)
        result_dict_raise_if_error(self.result)
        return self.result



from nose.tools import istest

from .compmake_test import CompmakeTest


def f():
    pass


def g(context):
    context.comp(f, job_id="ciao")  # this will become ciao-0


@istest
class TestDynamic3(CompmakeTest):
    def test_dynamic3(self):
        context = self.cc
        context.comp(f, job_id="ciao")
        self.assert_cmd_success("ls")
        self.assert_cmd_success("make")
        context.comp_dynamic(g, job_id="g")
        self.assert_cmd_success("make g")



from zuper_commons.text import get_length_on_screen, remove_escapes

from .terminal_size import get_screen_columns

__all__ = [
    "remove_escapes",
    "get_length_on_screen",
    "pad_to_screen",
    "pad_to_screen_length",
]

debug_padding = False


def pad_to_screen(s: str, pad=" ", last=None) -> str:
    """
        Pads a string to the terminal size.

        The string length is computed after removing shell escape sequences.
    """
    total_screen_length = get_screen_columns()

    if debug_padding:
        x = pad_to_screen_length(s, total_screen_length, pad="-", last="|", align_right=False)
    else:
        x = pad_to_screen_length(s, total_screen_length, pad=pad, last=last, align_right=False)

    return x


def check_not_bytes(x: str):
    if isinstance(x, bytes):
        msg = "You passed a bytes argument: %s" % x
        raise Exception(msg)


def pad_to_screen_length(s: str, desired_screen_length: int, pad=" ", last=None, align_right=False) -> str:
    assert len(pad) == 1, pad
    """
        Pads a string so that it will appear of the given size
        on the terminal.

        align_right: aligns right instead of left (default)
    """
    check_not_bytes(s)

    assert isinstance(desired_screen_length, int)
    # todo: assert pad = 1
    current_size = get_length_on_screen(s)

    if last is None:
        last = pad

    if current_size < desired_screen_length:
        nadd = desired_screen_length - current_size
        padding = pad * (nadd - 1)
        if align_right:
            s = last + padding + s
        else:
            s = s + padding + last

    if debug_padding:
        if current_size > desired_screen_length:
            T = "(cut)"
            s = s[:desired_screen_length] + T

    return s



import multiprocessing
import os
import random
import signal
import time
from multiprocessing import Queue
from multiprocessing.context import BaseContext
from typing import Dict, NewType, Set, Tuple

import psutil
from future.moves.queue import Empty
from psutil import NoSuchProcess

from compmake.events import broadcast_event, publish
from compmake.exceptions import MakeHostFailed
from compmake.jobs import Manager, parmake_job2_new_process

from zuper_commons.fs import make_sure_dir_exists
from zuper_commons.types import check_isinstance
from .parmake_job2_imp import parmake_job2
from .pmakesub import PmakeSub
from ...structures import CMJobID

__all__ = [
    "PmakeManager",
]

from ...ui.visualization import ui_warning


def killtree() -> None:
    # print('killing process tree')
    parent = psutil.Process(os.getpid())
    for child in parent.children(recursive=True):
        # print("child: %s"%child)
        try:
            child.kill()
        except NoSuchProcess:
            pass


SubName = NewType("SubName", str)


class PmakeManager(Manager):
    """
        Specialization of Manager for local multiprocessing, using
        an adhoc implementation of "pool" because of bugs of the
        Python 2.7 implementation of pool multiprocessing.
    """

    queues = {}

    event_queue: Queue
    event_queue_name: str
    subs: Dict[SubName, PmakeSub]
    sub_available: Set[SubName]
    sub_processing: Set[SubName]
    sub_aborted: Set[SubName]
    job2subname: Dict[CMJobID, SubName]
    max_num_processing: int

    def __init__(
        self,
        context,
        cq,
        num_processes: int,
        recurse: bool = False,
        new_process: bool = False,
        show_output: bool = False,
    ):
        Manager.__init__(self, context=context, cq=cq, recurse=recurse)
        self.num_processes = num_processes
        self.last_accepted = 0
        self.new_process = new_process
        self.show_output = show_output

        if new_process and show_output:
            msg = "Compmake does not yet support echoing stdout/stderr when jobs are run in a new process."
            ui_warning(context, msg)
        self.cleaned = False

    ctx: BaseContext

    def process_init(self) -> None:

        self.ctx = multiprocessing.get_context("fork")

        self.event_queue = self.ctx.Queue(1000)
        r = random.randint(0, 10000)
        self.event_queue_name = f"q-{id(self)}-{r}"

        PmakeManager.queues[self.event_queue_name] = self.event_queue

        # info(f"Starting {self.num_processes} processes queues = {PmakeManager.queues}")

        self.subs = {}  # name -> sub
        # available + processing + aborted = subs.keys
        self.sub_available = set()
        self.sub_processing = set()
        self.sub_aborted = set()

        db = self.context.get_compmake_db()
        storage = db.basepath  # XXX:
        logs = os.path.join(storage, "logs")

        # self.signal_queue = Queue()

        for i in range(self.num_processes):
            name = SubName(f"parmake_sub_{i:02d}")
            write_log = os.path.join(logs, f"{name}.log")
            make_sure_dir_exists(write_log)
            signal_token = name
            p = PmakeSub(
                name=name, signal_queue=None, signal_token=signal_token, write_log=write_log, ctx=self.ctx
            )
            self.subs[name] = p
        self.job2subname = {}
        # all are available
        self.sub_available.update(self.subs)

        self.max_num_processing = self.num_processes

    # XXX: boiler plate
    def get_resources_status(self) -> Dict[str, Tuple[bool, str]]:
        resource_available = {}

        assert len(self.sub_processing) == len(self.processing)

        if not self.sub_available:
            msg = f"already {len(self.sub_processing)} processing"
            if self.sub_aborted:
                msg += f" ({len(self.sub_aborted)} workers aborted)"
            resource_available["nproc"] = (False, msg)
            # this is enough to continue
            return resource_available
        else:
            resource_available["nproc"] = (True, "")

        return resource_available

    def can_accept_job(self, reasons_why_not: Dict) -> bool:
        if len(self.sub_available) == 0 and len(self.sub_processing) == 0:
            # all have failed
            msg = "All workers have aborted."
            raise MakeHostFailed(msg)

        resources = self.get_resources_status()
        some_missing = False
        for k, v in resources.items():
            if not v[0]:
                some_missing = True
                reasons_why_not[k] = v[1]
        if some_missing:
            return False
        return True

    def instance_job(self, job_id: CMJobID):
        publish(self.context, "worker-status", job_id=job_id, status="apply_async")
        assert len(self.sub_available) > 0
        name = sorted(self.sub_available)[0]
        self.sub_available.remove(name)
        assert not name in self.sub_processing
        self.sub_processing.add(name)
        sub = self.subs[name]

        self.job2subname[job_id] = name

        check_isinstance(job_id, str)
        if self.new_process:
            f = parmake_job2_new_process
            args = (job_id, self.context)

        else:
            f = parmake_job2
            logdir = os.path.join(self.context.get_compmake_db().basepath, "parmake_job2_logs")
            args = (job_id, self.context, self.event_queue_name, self.show_output, logdir)

        async_result = sub.apply_async(f, args)
        return async_result

    def event_check(self) -> None:
        if not self.show_output:
            return
        while True:
            try:
                event = self.event_queue.get(block=False)  # @UndefinedVariable
                event.kwargs["remote"] = True
                broadcast_event(self.context, event)
            except Empty:
                break

    def process_finished(self) -> None:
        if self.cleaned:
            return
        self.cleaned = True
        # print('process_finished()')

        self.event_queue.close()
        del PmakeManager.queues[self.event_queue_name]

        for name in self.sub_processing:
            self.subs[name].proc.terminate()

        for name in self.sub_available:
            self.subs[name].terminate()

        # XXX: in practice this never works well
        # if False:
        cps = os.environ.get("COVERAGE_PROCESS_START")
        if cps:
            self.log("Now waiting 5 seconds for coverage")
            time.sleep(10)
            self.log("Waited 5 seconds, now killing")
        else:
            self.log("Coverage not detected")

        # XXX: ... so we just kill them mercilessly
        if True:
            #  print('killing')
            for name in self.sub_processing:
                pid = self.subs[name].proc.pid
                os.kill(pid, signal.SIGKILL)
                # print('killed pid %s for %s' % (name, pid))
                # print('process_finished() finished')

        if False:
            timeout = 100
            for name in self.sub_available:
                print("joining %s" % name)
                self.subs[name].proc.join(timeout)

        killtree()
        # print('process_finished(): cleaned up')

    # Normal outcomes
    def job_failed(self, job_id: CMJobID, deleted_jobs):
        Manager.job_failed(self, job_id, deleted_jobs)
        self._clear(job_id)

    def job_succeeded(self, job_id: CMJobID):
        Manager.job_succeeded(self, job_id)
        self._clear(job_id)

    def _clear(self, job_id: CMJobID):
        assert job_id in self.job2subname
        name = self.job2subname[job_id]
        del self.job2subname[job_id]
        assert name in self.sub_processing
        assert name not in self.sub_available
        self.sub_processing.remove(name)
        self.sub_available.add(name)

    def host_failed(self, job_id: CMJobID):
        Manager.host_failed(self, job_id)

        assert job_id in self.job2subname
        name = self.job2subname[job_id]
        del self.job2subname[job_id]
        assert name in self.sub_processing
        assert name not in self.sub_available
        self.sub_processing.remove(name)

        # put in sub_aborted
        self.sub_aborted.add(name)

    def cleanup(self) -> None:
        self.process_finished()



from contextlib import contextmanager
import sys

import time


__all__ = [
    "TimeTrack",
]


class TimeTrack:
    def __init__(self, what=None):
        self.t0 = time.time()
        self.c0 = time.process_time()
        self.what = what

    def show(self, stream=sys.stdout, min_td=0.001):
        self.t1 = time.time()
        self.c1 = time.process_time()
        self.cd = self.c1 - self.c0
        self.td = self.t1 - self.t0

        if self.td < min_td:
            return

        msg = "wall %6.2fms clock %6.2fms" % (self.td * 1000, self.cd * 1000)
        if self.what:
            what = str(self.what)
            MAX = 120
            if len(what) > MAX:
                what = what[: (MAX - 3)] + "..."
            # msg = '%s - %s' % (msg, what)
            msg = "%s - %s" % (msg, what)
        # stream.write(msg)
        #         stream.write('\n')
        #         stream.flush()
        from compmake import logger

        logger.warning(msg)

    @staticmethod
    @contextmanager
    def measure(what=None, min_td=0.001):
        t = TimeTrack(what)
        yield
        t.show(min_td=min_td)

    @staticmethod
    def decorator(f):
        def wrapper(self, *args, **kwargs):
            sargs = ", ".join(["{0}".format(x) for x in args])
            if args and kwargs:
                sargs += ", "
            sargs += ", ".join([f"{k}={v!r}" for (k, v) in kwargs.items()])
            what = f"{f.__name__:>15}({sargs})"
            with TimeTrack.measure(what, min_td=0.1):
                return f(self, *args, **kwargs)

        return wrapper



from compmake import CompmakeBug
from compmake.jobs import AsyncResultInterface, make, Manager, parmake_job2_new_process
from compmake.jobs.result_dict import result_dict_check


# use_pympler = False
#
# if use_pympler:
#     from pympler import tracker  # @UnresolvedImport
#
#     tr = tracker.SummaryTracker()
# else:
from compmake.ui.visualization import ui_warning

tr = None


__all__ = [
    "ManagerLocal",
    "FakeAsync",
]


class ManagerLocal(Manager):
    """ Specialization of manager for local execution """

    # @contract(new_process="bool", echo="bool")
    def __init__(self, new_process: bool, echo: bool, *args, **kwargs):
        Manager.__init__(self, *args, **kwargs)
        self.new_process = new_process
        self.echo = echo

        if new_process and echo:
            msg = "Compmake does not yet support echoing stdout/stderr " "when jobs are run in a new process."
            ui_warning(self.context, msg)

    def can_accept_job(self, reasons):
        # only one job at a time
        if self.processing:
            reasons["cpu"] = "max 1 job"
            return False
        else:
            return True

    def instance_job(self, job_id):
        return FakeAsync(job_id, context=self.context, new_process=self.new_process, echo=self.echo)


class FakeAsync(AsyncResultInterface):
    def __init__(self, job_id, context, new_process, echo):
        self.job_id = job_id
        self.context = context
        self.new_process = new_process
        self.echo = echo

        self.told_you_ready = False

    def ready(self):
        if self.told_you_ready:
            msg = "Should not call ready() twice."
            raise CompmakeBug(msg)

        self.told_you_ready = True
        return True

    def get(self, timeout=0):
        if not self.told_you_ready:
            msg = "Should call get() only after ready()."
            raise CompmakeBug(msg)

        res = self._execute()
        result_dict_check(res)
        return res

    def _execute(self):
        if self.new_process:
            args = (self.job_id, self.context)
            return parmake_job2_new_process(args)
        else:
            # if use_pympler:
            #     tr.print_diff()

            return make(self.job_id, context=self.context, echo=self.echo)



import itertools
import os
import shutil
import time
import traceback
import warnings
from abc import ABCMeta, abstractmethod
from multiprocessing import TimeoutError
from typing import Any, Collection, Dict, List, Set

from compmake.constants import CompmakeConstants
from compmake.jobs.storage import db_job_add_dynamic_children, db_job_add_parent
from compmake.state import get_compmake_config

# from contracts import ContractsMeta, indent
from zuper_commons.fs import make_sure_dir_exists
from zuper_commons.text import indent
from .actions import mark_as_blocked
from .priority import compute_priorities
from .queries import direct_children, direct_parents
from .uptodate import CacheQueryDB
from .. import Context, StorageFilesystem
from ..events import publish
from ..exceptions import CompmakeBug, HostFailed, JobFailed, JobInterrupted
from ..jobs import assert_job_exists, get_job_cache, job_cache_exists, job_exists, job_userobject_exists
from ..jobs.actions_newprocess import result_dict_check
from ..structures import Cache, CMJobID, StateCode

__all__ = [
    "Manager",
    "AsyncResultInterface",
]


class AsyncResultInterface(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def ready(self):
        """ Returns True if it is ready (completed or failed). """

    @abstractmethod
    def get(self, timeout=0):
        """ Either:
            - returns a dictionary with fields:
                new_jobs: list of jobs created
                user_object_deps: ...
            or:
            - raises JobFailed
            - raises HostFailed
            - raises JobInterrupted
            - raises TimeoutError (not ready)
        """


class ManagerLog(object):
    # __metaclass__ = ContractsMeta

    def __init__(self, db):
        storage = os.path.abspath(db.basepath)
        logdir = os.path.join(storage, "logs")
        if os.path.exists(logdir):
            shutil.rmtree(logdir)
        log = os.path.join(logdir, "manager.log")
        # log = 'manager-%s.log' % sys.version
        # print('logging to %s' % log)
        make_sure_dir_exists(log)
        self.f = open(log, "w")

    def log(self, s: str, **kwargs: Any):

        for k in sorted(kwargs):
            v = kwargs[k]
            if isinstance(v, set):
                v = sorted(v)
            s += f"\n - {k:>15}: {v}"
        self.f.write(s)
        # print(s)
        self.f.write("\n")

        self.f.flush()


class Manager(ManagerLog):
    cq: CacheQueryDB
    context: Context
    db: StorageFilesystem
    targets: Set[CMJobID]
    """ top level"""
    all_targets: Set[CMJobID]
    deleted: Set[CMJobID]
    todo: Set[CMJobID]
    ready_todo: Set[CMJobID]
    processing: Set[CMJobID]
    processing2result: Dict[CMJobID, Any]

    # noinspection PyUnusedLocal
    def __init__(self, context, cq: CacheQueryDB, recurse: bool = False):
        self.context = context
        # self.cq = cq
        self.db = context.get_compmake_db()
        ManagerLog.__init__(self, db=self.db)

        self.recurse = recurse

        # top-level targets added by users
        self.targets = set()

        # top-level targets + all their dependencies
        self.all_targets = set()
        # some jobs might be deleted, they go here
        self.deleted = set()

        # all_targets - targets_delted = sum of rest:

        # a job is in exactly one of these states
        self.todo = set()
        # |
        # V
        self.ready_todo = set()
        # |
        # V
        # processing and processing2result have the same keys
        # (it's redundant)
        self.processing = set()
        # this hash contains  job_id -> async result
        self.processing2result = {}
        # |
        # V
        # final states
        self.done = set()
        self.failed = set()
        self.blocked = set()

        # contains job_id -> priority
        # computed by ``precompute_priorities()`` called by process()
        self.priorities = {}

        self.check_invariants()

    # ## Derived class interface
    def process_init(self):
        """ Called before processing """

    def process_finished(self):
        """ Called after successful processing (before cleanup) """

    @abstractmethod
    def can_accept_job(self, reasons: Dict[str, str]) -> bool:
        """ Return true if a new job can be accepted right away"""

    @abstractmethod
    def instance_job(self, job_id: CMJobID):
        """ Instances a job. """

    def cleanup(self):
        """ free up any resource, called wheter succesfull or not."""
        pass

    def next_job(self) -> CMJobID:
        """
            Returns one job from the ready_todo list
            Uses self.priorities to decide which job to use.
        """
        self.check_invariants()

        ordered = sorted(self.ready_todo, key=lambda job: self.priorities[job])
        best = ordered[-1]

        # print('choosing %s job %r' % (self.priorities[best], best))
        return best

    def add_targets(self, targets: Collection[CMJobID]):
        self.log("add_targets()", targets=targets)
        self.check_invariants()
        for t in targets:
            assert_job_exists(t, self.db)

            if t in self.processing:
                msg = f"Adding a job already in processing: {t!r}"
                raise CompmakeBug(msg)

            if t in self.targets:
                if t in self.ready_todo:
                    self.ready_todo.remove(t)
                if t in self.todo:
                    self.todo.remove(t)
                if t in self.all_targets:
                    self.all_targets.remove(t)
                if t in self.done:
                    self.done.remove(t)

        # self.targets contains all the top-level targets we were passed
        self.targets.update(targets)

        # logger.info('Checking dependencies...')
        cq = CacheQueryDB(self.db)
        # Note this would not work for recursive jobs
        # cq = self.cq
        targets_todo_plus_deps, targets_done, ready_todo = cq.list_todo_targets(targets)
        not_ready = targets_todo_plus_deps - ready_todo

        self.log(
            "computed todo",
            targets_todo_plus_deps=targets_todo_plus_deps,
            targets_done=targets_done,
            ready_todo=ready_todo,
            not_ready=not_ready,
        )

        self.log(f"targets_todo_plus_deps: {sorted(targets_todo_plus_deps)}")

        # print(' targets_todo_plus_deps: %s ' % targets_todo_plus_deps)
        # print('           targets_done: %s ' % targets_done)
        # print('             ready_todo: %s ' % ready_todo)
        # both done and todo jobs are added to self.all_targets

        # let's check the additional jobs exist
        for d in targets_todo_plus_deps - set(targets):
            if not job_exists(d, self.db):
                msg = "Adding job that does not exist: %r." % d
                raise CompmakeBug(msg)

        self.all_targets.update(targets_todo_plus_deps)
        self.all_targets.update(targets_done)

        # ok, careful here, there might be jobs that are
        # already in processing

        # XXX: we should clean the Cache of a job before making it
        self.done.update(targets_done - self.processing)

        todo_add = not_ready - self.processing
        self.todo.update(not_ready - self.processing)
        self.log("add_targets():adding to todo", todo_add=todo_add, todo=self.todo)
        ready_add = ready_todo - self.processing
        self.log("add_targets():adding to ready", ready=self.ready_todo, ready_add=ready_add)
        self.ready_todo.update(ready_add)
        # this is a quick fix but I'm sure more thought is to be given
        for a in ready_add:
            if a in self.todo:
                self.todo.remove(a)
        for a in todo_add:
            if a in self.ready_todo:
                self.ready_todo.remove(a)

        needs_priorities = self.todo | self.ready_todo
        misses_priorities = needs_priorities - set(self.priorities)
        new_priorities = compute_priorities(misses_priorities, cq=cq, priorities=self.priorities)
        self.priorities.update(new_priorities)

        self.check_invariants()

        self.log(
            "after add_targets()",
            processing=self.processing,
            ready_todo=self.ready_todo,
            todo=self.todo,
            done=self.done,
        )

    def instance_some_jobs(self):
        """
            Instances some of the jobs. Uses the
            functions can_accept_job(), next_job(), and ...

            Returns a dictionary of wait conditions.
        """
        self.check_invariants()

        n = 0
        reasons = {}
        while True:
            if not self.ready_todo:
                reasons["jobs"] = "no jobs ready"
                break

            if not self.can_accept_job(reasons):
                break

            # todo: add task priority
            job_id = self.next_job()
            assert job_id in self.ready_todo

            self.log("chosen next_job", job_id=job_id)

            self.start_job(job_id)
            n += 1

        #         print('cur %d Instanced %d, %s' % (len(self.processing2result), n, reasons))

        self.check_invariants()
        return reasons

    def _raise_bug(self, func, job_id):
        msg = f"{func}: Assumptions violated with job {job_id!r}."
        # msg += '\n' + self._get_situation_string()

        sets = [
            ("done", self.done),
            ("all_targets", self.all_targets),
            ("todo", self.todo),
            ("blocked", self.blocked),
            ("failed", self.failed),
            ("ready", self.ready_todo),
            ("processing", self.processing),
            ("proc2result", self.processing2result),
        ]

        for name, cont in sets:
            contained = job_id in cont
            msg += f"\n in {name:>15}? {contained}"

        raise CompmakeBug(msg)

    def start_job(self, job_id: CMJobID):
        self.log("start_job", job_id=job_id)
        self.check_invariants()
        if job_id not in self.ready_todo:
            self._raise_bug("start_job", job_id)

        publish(self.context, "manager-job-processing", job_id=job_id)
        self.ready_todo.remove(job_id)
        self.processing.add(job_id)
        self.processing2result[job_id] = self.instance_job(job_id)

        # This is for the simple case of local processing, where
        # the next line actually does something now
        self.publish_progress()

        self.check_invariants()

    def check_job_finished(self, job_id: CMJobID, assume_ready: bool = False):
        """
            Checks that the job finished succesfully or unsuccesfully.

            Returns True if that's the case.
            Captures HostFailed, JobFailed and returns True.

            Returns False if the job is still processing.

            Capture KeyboardInterrupt and raises JobInterrupted.

            Handles update of various sets.
        """
        self.log("check_job_finished", job_id=job_id)
        self.check_invariants()

        def bug():
            self._raise_bug("check_job_finished", job_id)

        if job_id not in self.processing:
            bug()

        async_result = self.processing2result[job_id]

        try:
            if not assume_ready and not async_result.ready():
                return False

            if assume_ready:
                timeout = 10
            else:
                timeout = 0

            result = async_result.get(timeout=timeout)
            # print('here result: %s' % result)
            result_dict_check(result)

            check_job_cache_state(job_id, states=[Cache.DONE], db=self.db)

            self.job_succeeded(job_id)
            self.check_invariants()

            self.check_job_finished_handle_result(job_id, result)
            self.check_invariants()
            # this will schedule the parents, so let's do it later

            self.check_invariants()

            return True

        except TimeoutError:
            if assume_ready:
                msg = "Got Timeout while assume_ready for %r" % job_id
                raise CompmakeBug(msg)
            # Result not ready yet
            return False
        except JobFailed as e:
            # it is the responsibility of the executer to mark_job_as_failed,
            # so we can check that
            check_job_cache_state(job_id, states=[Cache.FAILED], db=self.db)
            self.job_failed(job_id, deleted_jobs=e.deleted_jobs)

            publish(self.context, "job-failed", job_id=job_id, host="XXX", reason=e.reason, bt=e.bt)
            return True
        except HostFailed as e:
            # the execution has been interrupted, but not failed
            self.host_failed(job_id)
            publish(self.context, "manager-host-failed", host=e.host, job_id=job_id, reason=e.reason, bt=e.bt)
            return True
        except KeyboardInterrupt as e:
            # self.job_failed(job_id) # not sure
            # No, don't mark as failed
            # (even though knowing where it was interrupted was good)
            # XXX
            print(traceback.format_exc())
            raise JobInterrupted("Keyboard interrupt")

    def job_is_deleted(self, job_id: CMJobID):
        if job_exists(job_id, self.db):
            msg = f"Job {job_id!r} declared deleted still exists"
            raise CompmakeBug(msg)
        if job_id in self.all_targets:
            if job_id in self.todo:
                self.todo.remove(job_id)
            if job_id in self.ready_todo:
                self.ready_todo.remove(job_id)
            self.deleted.add(job_id)

    def check_job_finished_handle_result(self, job_id, result):
        self.check_invariants()
        self.log(
            "check_job_finished_handle_result",
            job_id=job_id,
            new_jobs=result["new_jobs"],
            user_object_deps=result["user_object_deps"],
            deleted_jobs=result["deleted_jobs"],
        )

        new_jobs = result["new_jobs"]
        deleted_jobs = result["deleted_jobs"]
        # self.log('deleted jobs: %r' % list(deleted_jobs))
        for _ in deleted_jobs:
            self.job_is_deleted(_)
        # print('Job %r generated %r' % (job_id, new_jobs))

        # Update the child->parent relation
        # self._update_parents_relation(new_jobs)

        # Job succeeded? we can check in the DB
        check_job_cache_state(job_id=job_id, db=self.db, states=[Cache.DONE])

        # print('job %r succeeded' % job_id)
        self.check_invariants()

        # Check if the result of this job contains references
        # to other jobs
        deps = result["user_object_deps"]
        if deps:
            # print('Job %r results contain references to jobs: %s'
            # % (job_id, deps))

            # We first add extra dependencies to all those jobs
            jobs_depending_on_this = direct_parents(job_id, self.db)
            # print('need to update %s' % jobs_depending_on_this)
            for parent in jobs_depending_on_this:
                db_job_add_dynamic_children(job_id=parent, children=deps, returned_by=job_id, db=self.db)

                # also add inverse relation
                for d in deps:
                    self.log("updating dep", job_id=job_id, parent=parent, d=d)
                    db_job_add_parent(job_id=d, parent=parent, db=self.db)

            for parent in jobs_depending_on_this:
                self.log("rescheduling parent", job_id=job_id, parent=parent)
                # print(' its parent %r' % parent)
                if parent in self.all_targets:
                    # print('was also in targets')
                    # Remove it from the "ready_todo_list"
                    if parent in self.processing2result:
                        msg = f"parent {job_id} of {parent} is already processing?"
                        raise CompmakeBug(msg)

                    if parent in self.done:
                        msg = f" parent {job_id} of {parent} is already done?"
                        warnings.warn("not sure of this...")
                        # raise CompmakeBug(msg)#

                    self.all_targets.remove(parent)
                    if parent in self.failed:
                        self.failed.remove(parent)
                    if parent in self.blocked:
                        self.blocked.remove(parent)
                    if parent in self.done:
                        self.done.remove(parent)
                    if parent in self.ready_todo:
                        self.ready_todo.remove(parent)
                    if parent in self.todo:
                        self.todo.remove(parent)
                    self.check_invariants()

                    self.add_targets([parent])
                    self.check_invariants()

        if self.recurse:
            # print('adding targets %s' % new_jobs)
            cocher = set()
            for j in new_jobs:
                if j in self.all_targets:
                    # msg = ('Warning, job %r generated %r which was '
                    # 'already a target. I will not re-add it to the
                    # queue. '
                    # % (job_id, j))
                    # print(msg)
                    pass
                else:
                    cocher.add(j)
            if cocher:
                self.add_targets(cocher)

        self.check_invariants()

    def host_failed(self, job_id):
        self.log("host_failed", job_id=job_id)
        self.check_invariants()

        # from compmake.ui.visualization import error
        # error('Host failed, rescheduling job %r.' % job_id)
        self.processing.remove(job_id)
        del self.processing2result[job_id]
        # rescheduling
        self.ready_todo.add(job_id)

        self.publish_progress()

        self.check_invariants()

    def job_failed(self, job_id: CMJobID, deleted_jobs):
        """ The specified job has failed. Update the structures,
            mark any parent as failed as well. """
        self.log("job_failed", job_id=job_id, deleted_jobs=deleted_jobs)
        self.check_invariants()
        assert job_id in self.processing

        for _ in deleted_jobs:
            self.job_is_deleted(_)

        self.failed.add(job_id)
        self.processing.remove(job_id)
        del self.processing2result[job_id]

        self.check_invariants()

        publish(self.context, "manager-job-failed", job_id=job_id)

        # TODO: more efficient query
        # parent_jobs = set(parents(job_id, db=self.db))
        from compmake.jobs.uptodate import direct_uptodate_deps_inverse_closure

        parent_jobs = direct_uptodate_deps_inverse_closure(job_id, db=self.db)

        parents_todo = set(self.todo & parent_jobs)
        for p in parents_todo:
            if p not in self.blocked:
                publish(self.context, "manager-job-blocked", job_id=p, blocking_job_id=job_id)

                mark_as_blocked(p, dependency=job_id, db=self.db)
                self.todo.remove(p)
                self.blocked.add(p)

        self.publish_progress()
        self.check_invariants()

    def job_succeeded(self, job_id: CMJobID):
        """ Mark the specified job as succeeded. Update the structures,
            mark any parents which are ready as ready_todo. """
        self.log("job_succeeded", job_id=job_id)
        self.check_invariants()
        publish(self.context, "manager-job-done", job_id=job_id)
        assert job_id in self.processing

        self.processing.remove(job_id)
        del self.processing2result[job_id]
        self.done.add(job_id)

        # parent_jobs = set(direct_parents(job_id, db=self.db))
        from compmake.jobs.uptodate import direct_uptodate_deps_inverse

        parent_jobs = direct_uptodate_deps_inverse(job_id, db=self.db)
        cq = CacheQueryDB(self.db)

        parents_todo = set(self.todo & parent_jobs)
        self.log("considering parents", parents_todo=parents_todo)
        for opportunity in parents_todo:
            # print('parent %r in todo' % (opportunity))
            if opportunity in self.processing:
                msg = f"Parent {opportunity!r} of {job_id!r} already processing"
                if CompmakeConstants.try_recover:
                    print(msg)
                    continue
                else:
                    raise CompmakeBug(msg)
            assert opportunity not in self.processing

            self.log("considering opportuniny", opportunity=opportunity, job_id=job_id)

            its_children = direct_children(opportunity, db=self.db)
            # print('its children: %r' % its_children)

            for child in its_children:
                # If child is part of all_targets, check that it is done
                # otherwise check that it is done by the DB.
                if child in self.all_targets:
                    if not child in self.done:
                        self.log("parent still waiting another child", opportunity=opportunity, child=child)
                        # logger.info('parent %r still waiting on %r' %
                        # (opportunity, child))
                        # still some dependency left
                        break
                else:
                    up, _, _ = cq.up_to_date(child)
                    if not up:
                        # print('The child %s is not up_to_date' % child)
                        break

            else:
                # print('parent %r is now ready' % (opportunity))
                self.log("parent is ready", opportunity=opportunity)
                self.todo.remove(opportunity)
                publish(self.context, "manager-job-ready", job_id=opportunity)
                self.ready_todo.add(opportunity)

        self.check_invariants()
        self.publish_progress()

    def event_check(self):
        pass

    def check_any_finished(self):
        """
            Checks that any of the jobs finished.

            Returns True if something finished (either success or failure).
            Returns False if something finished unseccesfully.
        """
        # We make a copy because processing is updated during the loop
        received = False
        for job_id in self.processing.copy():
            received = received or self.check_job_finished(job_id)
            self.check_invariants()
        return received

    def loop_until_something_finishes(self):
        self.check_invariants()

        manager_wait = get_compmake_config("manager_wait")

        # TODO: this should be loop_a_bit_and_then_let's try to instantiate
        # jobs in the ready queue
        for _ in range(10):  # XXX
            received = self.check_any_finished()

            if received:
                break
            else:
                publish(self.context, "manager-loop", processing=list(self.processing))
                time.sleep(manager_wait)  # TODO: make param

            # Process events
            self.event_check()
            self.check_invariants()

    def process(self):
        """ Start processing jobs. """
        # logger.info('Started job manager with %d jobs.' % (len(self.todo)))
        self.check_invariants()

        if not self.todo and not self.ready_todo:
            publish(
                self.context,
                "manager-succeeded",
                nothing_to_do=True,
                targets=self.targets,
                done=self.done,
                all_targets=self.all_targets,
                todo=self.todo,
                failed=self.failed,
                blocked=self.blocked,
                ready=self.ready_todo,
                processing=self.processing,
            )
            return True

        publish(self.context, "manager-phase", phase="init")
        self.process_init()

        publish(self.context, "manager-phase", phase="loop")
        try:
            i = 0
            while self.todo or self.ready_todo or self.processing:
                self.log(indent(self._get_situation_string(), f"{i}: "))
                i += 1
                self.check_invariants()
                # either something ready to do, or something doing
                # otherwise, we are completely blocked
                if (not self.ready_todo) and (not self.processing):
                    msg = (
                        "Nothing ready to do, and nothing cooking. "
                        "This probably means that the Compmake job "
                        "database was inconsistent. "
                        "This might happen if the job creation is "
                        'interrupted. Use the command "check-consistency" '
                        "to check the database consistency.\n" + self._get_situation_string()
                    )
                    raise CompmakeBug(msg)

                self.publish_progress()
                waiting_on = self.instance_some_jobs()
                # self.publish_progress()

                publish(self.context, "manager-wait", reasons=waiting_on)

                if self.ready_todo and not self.processing:
                    # We time out as there are no resources
                    publish(self.context, "manager-phase", phase="wait")

                self.loop_until_something_finishes()
                self.check_invariants()
            self.log(indent(self._get_situation_string(), "ending: "))

            # end while
            assert not self.todo
            assert not self.ready_todo
            assert not self.processing
            self.check_invariants()

            self.publish_progress()

            self.process_finished()

            publish(
                self.context,
                "manager-succeeded",
                nothing_to_do=False,
                targets=self.targets,
                done=self.done,
                all_targets=self.all_targets,
                todo=self.todo,
                failed=self.failed,
                ready=self.ready_todo,
                blocked=self.blocked,
                processing=self.processing,
            )

            return True

        except JobInterrupted as e:

            from compmake.ui.visualization import ui_error

            ui_error(self.context, f"Received JobInterrupted: {e}")
            raise
        except KeyboardInterrupt:
            raise KeyboardInterrupt("Manager interrupted.") from None
        finally:
            self.cleanup()

    def publish_progress(self):
        publish(
            self.context,
            "manager-progress",
            targets=self.targets,
            done=self.done,
            all_targets=self.all_targets,
            todo=self.todo,
            blocked=self.blocked,
            failed=self.failed,
            ready=self.ready_todo,
            processing=self.processing,
            deleted=self.deleted,
        )

    def _get_situation_string(self):
        """ Returns a string summarizing the current situation """
        lists = dict(
            done=self.done,
            all_targets=self.all_targets,
            todo=self.todo,
            blocked=self.blocked,
            failed=self.failed,
            ready=self.ready_todo,
            deleted=self.deleted,
            processing=self.processing,
        )
        s = ""
        for t in sorted(lists):
            jobs = lists[t]
            s += f"- {t:>12}: {len(jobs)}\n"

        # if False:
        s += "\n In more details:"
        for t in sorted(lists):
            jobs = lists[t]
            if not jobs:
                s += f"\n- {t:>12}: -"
            elif len(jobs) < 20:
                s += f"\n- {t:>12}: {len(jobs)} {sorted(jobs)}"

        return s

    def check_invariants(self):
        if not CompmakeConstants.debug_check_invariants:
            return
        lists = dict(
            done=self.done,
            all_targets=self.all_targets,
            todo=self.todo,
            blocked=self.blocked,
            failed=self.failed,
            ready_todo=self.ready_todo,
            processing=self.processing,
            deleted=self.deleted,
        )

        def empty_intersection(a, b):
            inter = lists[a] & lists[b]
            if inter:
                msg = f"There should be empty intersection in {a!r} and {b!r}"
                msg += f" but found {inter}"

                st = self._get_situation_string()
                if len(st) < 500:
                    msg += "\n" + st

                raise CompmakeBug(msg)

        def partition(sets, result):
            S = set()
            for s in sets:
                S |= lists[s]

            if S != lists[result]:
                msg = "These two sets should be the same:\n"
                msg += " %s = %s\n" % (" + ".join(list(sets)), result)
                msg += f" first = {S}\n"
                msg += f" second = {lists[result]}\n"
                msg += f" first-second = {S - lists[result]}\n"
                msg += f" second-first = {lists[result] - S}\n"

                st = self._get_situation_string()
                if len(st) < 500:
                    msg += "\n" + st

                raise CompmakeBug(msg)

            for a, b in itertools.product(sets, sets):
                if a == b:
                    continue
                empty_intersection(a, b)

        partition(["done", "failed", "blocked", "todo", "ready_todo", "processing", "deleted"], "all_targets")

        if False:

            for job_id in self.done:
                if not job_exists(job_id, self.db):
                    raise CompmakeBug("job %r in done does not exist" % job_id)

            for job_id in self.todo:
                if not job_exists(job_id, self.db):
                    raise CompmakeBug("job %r in todo does not exist" % job_id)

            for job_id in self.failed:
                if not job_exists(job_id, self.db):
                    raise CompmakeBug("job %r in failed does not exist" % job_id)

            for job_id in self.blocked:
                if not job_exists(job_id, self.db):
                    raise CompmakeBug("job %r in blocked does not exist" % job_id)


def check_job_cache_state(job_id: CMJobID, states: List[StateCode], db):
    """ Raises CompmakeBug if the job is not marked as done. """
    if not CompmakeConstants.extra_checks_job_states:  # XXX: extra check
        return

    if not job_cache_exists(job_id, db):
        msg = f"The job {job_id!r} was reported as done/failed but no record of it was found."
        raise CompmakeBug(msg)
    else:
        cache = get_job_cache(job_id, db)
        if not cache.state in states:
            possible = [Cache.state2desc[s] for s in states]
            found = Cache.state2desc[cache.state]
            msg = f"Wrong state for {job_id!r}: {found} instead of {possible!r} "
            raise CompmakeBug(msg)

        if cache.state == Cache.DONE:
            if not job_userobject_exists(job_id, db):
                msg = f"Job {job_id!r} marked as DONE but no userobject exists"
                raise CompmakeBug(msg)


#
# if False:
#     def clean_other_jobs(self, job_id, new_jobs):
#         """ job_id has finished and the jobs in new_jobs have been
#             generated. We should look in the DB if in the past
#             it had generated other jobs and delete them """
#         # print('cleaning other jobs after %r generated %r' % (job_id,
#         # new_jobs))
#         db = self.db
#         extra = []
#         # XXX: slow
#         for g in all_jobs(db=db):
#             if get_job(g, db=db).defined_by[-1] == job_id:
#                 if not g in new_jobs:
#                     extra.append(g)
#
#         for g in extra:
#             if g in self.processing:
#                 print(
#                     'a mess - cannot eliminate job %s because processing' % g)
#             else:
#                 if g in self.targets:
#                     # print('removing job %r which was an explicit target' % g)
#                     self.targets.remove(g)
#                 if g in self.all_targets:
#                     self.all_targets.remove(g)
#                 if g in self.todo:
#                     self.todo.remove(g)
#                 if g in self.ready_todo:
#                     self.ready_todo.remove(g)
#                 if g in self.ready_todo:
#                     self.todo.remove(g)
#                 if g in self.failed:
#                     self.failed.remove(g)
#                 if g in self.blocked:
#                     self.blocked.remove(g)
#
#             # print('Erasing previously generated job %r (%s) removed.' % (
#             # g, job.defined_by))
#             delete_all_job_data(g, db=db)
#
#             # clean dependencies as well
#             self.clean_other_jobs(g, [])


def check_job_cache_says_failed(job_id, db, e):
    """ Raises CompmakeBug if the job is not marked as failed. """
    if not job_cache_exists(job_id, db):
        msg = f"The job {job_id!r} was reported as failed but no record of it was found."
        # msg += "\n" + "JobFailed exception:"
        # msg += "\n" + indent(str(e), "| ")
        raise CompmakeBug(msg, e=str(e))
    else:
        cache = get_job_cache(job_id, db)
        if not cache.state == Cache.FAILED:
            msg = f"The job {job_id!r} was reported as failed but it was not marked as such in the DB."
            msg += f"\n seen state: {Cache.state2desc[cache.state]} "
            # msg += "\n" + "JobFailed exception:"
            # msg += "\n" + indent(str(e), "| ")
            raise CompmakeBug(msg, e=str(e))


#
#
# def clean_other_jobs_distributed(db, job_id, new_jobs, recurse=False):
# """ job_id has finished and the jobs in new_jobs have been
# generated. We should look in the DB if in the past
# it had generated other jobs and delete them """
# #print('cleaning other jobs after %r generated %r' % (job_id, new_jobs))
#     extra = []
#     # XXX: slow
#     for g in all_jobs(db=db):
#         try:
#             job = get_job(g, db)
#         except:
#             # race condition
#             continue
#
#         if job.defined_by[-1] == job_id:
#             if not g in new_jobs:
#                 extra.append(g)
#
#         delete_all_job_data(g, db=db)
#
#         # clean dependencies as well
#         if recurse:
#             clean_other_jobs_distributed(db, g, [])



import sys

__all__ = [
    "setproctitle",
]

try:
    from setproctitle import setproctitle  # @UnresolvedImport @UnusedImport
except:
    msg = 'compmake can make use of the package "setproctitle". ' "Please install it.\n"
    sys.stderr.write(msg)

    def setproctitle(x):
        """ emulation of the setproctitle interface """
        pass



from compmake.unittests.compmake_test import CompmakeTest
from nose.tools import istest


def always():
    print("always()")
    pass


def other():
    print("other()")
    pass


def dep():
    res = TestDynamic9.define_other
    print("other() returns %s" % res)
    return res


def fd(context, dep):
    context.comp(always)
    print("fd() sees dep=%s" % dep)
    if dep:
        context.comp(other)


def mockup9(context):
    depres = context.comp(dep)
    context.comp_dynamic(fd, dep=depres)


@istest
class TestDynamic9(CompmakeTest):
    define_other = True

    def test_dynamic9(self):
        """ Re-execution creates more jobs.  """
        mockup9(self.cc)

        self.assert_cmd_success("config echo 1")
        self.assert_cmd_success("config echo_stdout 1")
        self.assert_cmd_success("config echo_stderr 1")
        # self.assert_cmd_success('config console_status 1')
        # run it
        TestDynamic9.define_other = True  # returned by dep
        self.assert_cmd_success("make recurse=1")
        self.assert_cmd_success("stats")
        # we have 4 jobs defined
        self.assertJobsEqual("all", ["fd", "fd-always", "fd-other", "dep"])
        # clean and remake fd
        TestDynamic9.define_other = False  # returned by dep
        # clean dep
        self.assert_cmd_success("remake dep")
        self.assert_cmd_success("stats")

        # now all jobs are done
        self.assertJobsEqual("done", ["fd", "fd-always", "fd-other", "dep"])
        # but fd is not up to date
        self.assert_job_uptodate("fd", False)

        self.assert_cmd_success("make echo=1")  # remaking
        # now the "other" job should disappear
        self.assert_cmd_success("stats")
        self.assertJobsEqual("all", ["fd", "fd-always", "dep"])



from compmake.ui.helpers import GENERAL, ui_command


@ui_command(section=GENERAL)
def clear():  # @ReservedAssignment
    """ Clear the screen. """
    import sys

    sys.stderr.write("\x1b[2J\x1b[H")
    # http://stackoverflow.com/questions/2084508/clear-terminal-in-python
    sys.stderr.write("\n")



#
# @ui_command(section=COMMANDS_CLUSTER, dbchange=True)
# def clustmake(job_list, context, cq):
# ''' (experimental) Cluster equivalent of "make". '''
#     # job_list = list(job_list) # don't ask me why XXX
#     job_list = [x for x in job_list]
#
#     if not job_list:
#         db = context.get_compmake_db()
#         job_list = list(top_targets(db))
#
#     cluster_conf = get_compmake_config('cluster_conf')
#
#     if not os.path.exists(cluster_conf):
#         msg = ('Configuration file %r does not exist.' % cluster_conf)
#         raise UserError(msg)
#
#     hosts = parse_yaml_configuration(open(cluster_conf))
#     manager = ClusterManager(hosts=hosts, context=context, cq=cq)
#     manager.add_targets(job_list)
#     manager.process()
#     return raise_error_if_manager_failed(manager)



from .compmake_test import CompmakeTest
from nose.tools import istest


def job_success(*args, **kwargs):
    pass


def job_failure(*args, **kwargs):
    assert False, "asserting false"


@istest
class TestAssertion(CompmakeTest):
    def mySetUp(self):
        pass

    def testAssertion1(self):
        for i in range(10):
            self.comp(job_failure, job_id="fail%d" % i)

        def run():
            self.cc.batch_command("parmake n=2")

        self.assertMakeFailed(run, nfailed=10, nblocked=0)



"""
    These are all the functionalities that build on the API. All visualization
    stuff is built in as a plugin.

"""

from . import (
    backend_local,
    backend_pmake,
    backend_sge,
    backend_ssh_cluster,
    clear_imp,
    commands_status,
    console_banners,
    console_output,
    console_status,
    credits,
    debug_priority,
    details,
    details_why,
    dump,
    event_debugger,
    gantt,
    graph,
    graph_animation_imp,
    job_definition_status,
    list_jobs_imp,
    reload_module,
    sanity_check,
    stats,  # handle *before* console_status; handle *after*  console_output
)

# Useful for debugging events
# TODO: mail, html_status
# TODO: add "plugin X, automatically loading a plugin"



import traceback

from compmake import CompmakeGlobalState, logger
from compmake.context import Context
from zuper_commons.text import indent
from zuper_commons.fs import make_sure_dir_exists
from zuper_commons.types import ZValueError
from .registered_events import compmake_registered_events
from .structures import Event
from ..exceptions import CompmakeException
from ..utils import wildcard_to_regexp

__all__ = [
    "broadcast_event",
    "remove_all_handlers",
    "register_fallback_handler",
    "register_handler",
    "publish",
]


def remove_all_handlers():
    """
        Removes all event handlers. Useful when
        events must not be processed locally but routed
        to the original process somewhere else.
    """
    CompmakeGlobalState.EventHandlers.handlers = {}
    CompmakeGlobalState.EventHandlers.fallback = []


def register_fallback_handler(handler):
    """
        Registers an handler who is going to be called when no other handler
        can deal with an event. Useful to see if we are ignoring some event.
    """
    CompmakeGlobalState.EventHandlers.fallback.append(handler)


import inspect

# TODO: make decorator
def register_handler(event_name: str, handler):
    """
        Registers an handler with an event name.
        The event name might contain asterisks. "*" matches all.
    """

    spec = inspect.getfullargspec(handler)
    args = set(spec.args)
    possible_args = {"event", "context", "self"}
    # to be valid
    if not (args.issubset(possible_args)):
        #     if not 'context' in args and 'event' in args:
        msg = "Function is not valid event handler"
        raise ZValueError(msg, handler=handler, args=spec)
    handlers = CompmakeGlobalState.EventHandlers.handlers

    if event_name.find("*") > -1:
        regexp = wildcard_to_regexp(event_name)

        for event in compmake_registered_events.keys():
            if regexp.match(event):
                register_handler(event, handler)

    else:
        if event_name not in handlers:
            handlers[event_name] = []
        handlers[event_name].append(handler)


def publish(context: Context, event_name: str, **kwargs):
    """ Publishes an event. Checks that it is registered and with the right
        attributes. Then it is passed to broadcast_event(). """
    if event_name not in compmake_registered_events:
        msg = "Event %r not registered" % event_name
        logger.error(msg)
        raise CompmakeException(msg)
    spec = compmake_registered_events[event_name]
    for key in kwargs.keys():
        if key not in spec.attrs:
            msg = "Passed attribute %r for event type %r but only found " "attributes %s." % (
                key,
                event_name,
                spec.attrs,
            )
            logger.error(msg)
            raise CompmakeException(msg)
    event = Event(event_name, **kwargs)
    broadcast_event(context, event)


# @contract(context=Context, event=Event)
def broadcast_event(context: Context, event: Event):

    all_handlers = CompmakeGlobalState.EventHandlers.handlers

    handlers = all_handlers.get(event.name, [])
    if handlers:
        for handler in handlers:
            spec = inspect.getfullargspec(handler)
            try:
                kwargs = {}
                if "event" in spec.args:
                    kwargs["event"] = event
                if "context" in spec.args:
                    kwargs["context"] = context
                handler(**kwargs)
                # TODO: do not catch interrupted, etc.
            except KeyboardInterrupt:
                raise
            except BaseException:
                try:
                    msg = [
                        "compmake BUG: Error in event handler.",
                        "  event: %s" % event.name,
                        "handler: %s" % handler,
                        " kwargs: %s" % list(event.kwargs.keys()),
                        "     bt: ",
                        indent(traceback.format_exc(), "| "),
                    ]
                    msg = "\n".join(msg)
                    CompmakeGlobalState.original_stderr.write(msg)
                except:
                    pass
    else:
        for handler in CompmakeGlobalState.EventHandlers.fallback:
            handler(context=context, event=event)


import os


def get_events_log_file(db):
    storage = os.path.abspath(db.basepath)
    logdir = os.path.join(storage, "events")
    lf = os.path.join(logdir, "events.log")
    make_sure_dir_exists(lf)
    if not os.path.exists(lf):
        with open(lf, "w") as f:
            f.write("first.\n")
    return lf


def handle_event_logs(context: Context, event):
    db = context.compmake_db
    lf = get_events_log_file(db)
    with open(lf, "a") as f:
        f.write(str(event) + "\n")


register_handler("*", handle_event_logs)



from nose.tools import istest

from .compmake_test import CompmakeTest
from compmake import compmake_execution_stats
from zuper_commons.types import check_isinstance
from compmake.jobs.dependencies import get_job_userobject_resolved
from ..structures import CMJobID


def ff(*args):
    return


def gg(context):
    return context.comp(ff)


def hh(context):
    return context.comp_dynamic(gg)


@istest
class TestExecutionStats(CompmakeTest):
    def test_execution_stats(self):
        # schedule some commands
        res = self.cc.comp_dynamic(gg)

        myjobid = CMJobID("myjobid")
        compmake_execution_stats(self.cc, res, use_job_id=myjobid)
        self.assert_cmd_success("make recurse=1")

        res = get_job_userobject_resolved(myjobid, db=self.db)
        check_result(res)

        _ = res["cpu_time"]
        _ = res["wall_time"]

        print(res)
        self.assertEqual(res["jobs"], {"gg-ff", "gg"})

    def test_execution_stats2(self):
        # schedule some commands
        res = self.cc.comp_dynamic(hh)

        myjobid = "myjobid"
        compmake_execution_stats(self.cc, res, use_job_id=myjobid)
        self.assert_cmd_success("make recurse=1")
        self.assert_cmd_success("ls")

        res = get_job_userobject_resolved(myjobid, db=self.db)
        check_result(res)

        print(res)

        self.assertEqual(res["jobs"], set(["hh-gg-ff", "hh-gg", "hh"]))


def check_result(res):
    check_isinstance(res, dict)
    _ = res["cpu_time"]
    _ = res["wall_time"]
    _ = res["jobs"]



from typing import Collection

from compmake.constants import DefaultsToConfig
from compmake.events import publish
from compmake.jobs import CacheQueryDB, CMJobID, top_targets
from compmake.jobs.actions import mark_to_remake
from compmake.ui import ACTIONS, ask_if_sure_remake, raise_error_if_manager_failed, ui_command
from .pmake_manager import PmakeManager

__all__ = [
    "parmake",
    "rparmake",
    "parremake",
]


@ui_command(section=ACTIONS, dbchange=True)
def parmake(
    job_list,
    context,
    cq,
    n: int = DefaultsToConfig("max_parallel_jobs"),
    recurse: bool = DefaultsToConfig("recurse"),
    new_process: bool = DefaultsToConfig("new_process"),
    echo: bool = DefaultsToConfig("echo"),
):
    """
        Parallel equivalent of make.

        Uses multiprocessing.Process as a backend and a Python queue to
        communicate with the workers.

        Options:

          parmake n=10             Uses 10 workers
          parmake recurse=1        Recursive make: put generated jobs in the
          queue.
          parmake new_process=1    Run the jobs in a new Python process.
          parmake echo=1           Shows the output of the jobs. This might
          slow down everything.

          parmake new_process=1 echo=1   Not supported yet.

    """

    publish(context, "parmake-status", status="Obtaining job list")
    job_list = list(job_list)

    db = context.get_compmake_db()
    if not job_list:
        # XXX
        job_list = list(top_targets(db=db))

    publish(context, "parmake-status", status="Starting multiprocessing manager (forking)")
    manager = PmakeManager(
        num_processes=n, context=context, cq=cq, recurse=recurse, new_process=new_process, show_output=echo
    )

    publish(context, "parmake-status", status="Adding %d targets." % len(job_list))
    manager.add_targets(job_list)

    publish(context, "parmake-status", status="Processing")
    manager.process()

    return raise_error_if_manager_failed(manager)


@ui_command(section=ACTIONS, dbchange=True)
def parremake(
    non_empty_job_list,
    context,
    cq,
    n: int = DefaultsToConfig("max_parallel_jobs"),
    recurse: bool = DefaultsToConfig("recurse"),
    new_process: bool = DefaultsToConfig("new_process"),
    echo: bool = DefaultsToConfig("echo"),
):
    """
        Parallel equivalent of "remake".
    """
    # TODO: test this
    db = context.get_compmake_db()
    non_empty_job_list = list(non_empty_job_list)

    if not ask_if_sure_remake(non_empty_job_list):
        return

    for job in non_empty_job_list:
        mark_to_remake(job, db=db)

    manager = PmakeManager(
        num_processes=n, context=context, cq=cq, recurse=recurse, new_process=new_process, show_output=echo
    )
    manager.add_targets(non_empty_job_list)
    manager.process()
    return raise_error_if_manager_failed(manager)


@ui_command(section=ACTIONS, dbchange=True)
def rparmake(
    job_list: Collection[CMJobID],
    context,
    cq: CacheQueryDB,
    n: int = DefaultsToConfig("max_parallel_jobs"),
    new_process: bool = DefaultsToConfig("new_process"),
    echo: bool = DefaultsToConfig("echo"),
):
    """ Shortcut to parmake with default recurse = True. """
    return parmake(
        job_list=job_list, context=context, cq=cq, n=n, new_process=new_process, echo=echo, recurse=True
    )



import os
from collections import defaultdict
from typing import Collection

from compmake.exceptions import UserError
from compmake.jobs import CacheQueryDB, CMJobID, top_targets
from compmake.jobs.queries import definition_closure
from compmake.structures import Cache
from compmake.ui import COMMANDS_ADVANCED, Context, ui_command
from compmake.ui.visualization import ui_info


@ui_command(section=COMMANDS_ADVANCED)
def graph(
    job_list: Collection[CMJobID],
    context: Context,
    filename: str = "compmake-graph",
    filter: str = "dot",
    format: str = "png",  # @ReservedAssignment
    label: str = "id",
    color: bool = True,
    cluster: bool = False,
    processing: Collection[CMJobID] = None,
):
    """

        Creates a graph of the given targets and dependencies.

        Usage:

            @: graph filename=filename label=[id,function,none] color=[0|1] format=png filter=[dot|circo|...]

        Options:

            filename:  name of generated filename in the dot format
            label='none','id','function'
            color=[0|1]: whether to color the nodes
            filter=[dot,circo,twopi,...]  which algorithm to use to arrange
                       the nodes. The best choice depends on
                       the topology of your
                       computation. The default is 'dot'
                       (hierarchy top-bottom).
            format=[png,...]  The output file format.
    """
    processing = processing or []
    possible = ["none", "id", "function"]
    if not label in possible:
        msg = f"Invalid label method {label!r} not in {possible!r}."
        raise ValueError(msg)

    db = context.get_compmake_db()
    if not job_list:
        job_list = list(top_targets(db))

    print(f"jobs: {job_list}")
    print(f"processing: {processing}")
    print("Importing gvgen")

    try:
        #        import gvgen
        pass
    except:
        gvgen_url = "https://github.com/stricaud/gvgen"
        msg = 'To use the "graph" command you have to install the "gvgen" ' "package from %s" % gvgen_url
        raise UserError(msg)

    print("Getting all jobs in tree")

    cq = CacheQueryDB(db)
    job_list = set(job_list)
    # all the dependencies
    job_list.update(cq.tree(job_list))

    # plus all the jobs that were defined by them
    job_list.update(definition_closure(job_list, db))

    job_list = set(job_list)

    #     print('closure: %s' % sorted(job_list))

    if cluster:
        ggraph = create_graph2_clusters(cq, job_list, label=label, color=color, processing=processing)
    else:
        ggraph = create_graph1(cq, job_list, label=label, color=color, processing=processing)
    print(f"Writing graph on {filename!r}.")
    # TODO: add check?

    with open(filename, "w") as f:
        ggraph.dot(f)

    print("Running rendering")
    output = filename + "." + format
    cmd_line = f"{filter} {filename} -T{format} -o{output}"
    print(f"  {cmd_line}")
    try:
        os.system(cmd_line)
    except:
        msg = "Could not run dot (cmdline={cmd_line!r}). Make sure graphviz is installed."
        raise UserError(msg)  # XXX maybe not UserError

    ui_info(context, f"Written output on files {filename}, {output}.")


def get_color_for(x: CMJobID, cq: CacheQueryDB, processing: Collection[CMJobID]):
    cache = cq.get_job_cache(x)

    # state2color = {
    #     Cache.NOT_STARTED: "grey",
    #     #         Cache.IN_PROGRESS: 'yellow',
    #     Cache.FAILED: "red",
    #     Cache.DONE: "green",
    #     Cache.BLOCKED: "brown",
    # }

    state = cache.state

    if x in processing and state != Cache.PROCESSING:
        state = Cache.PROCESSING

    return Cache.state2color[state]


def get_node_label(cq: CacheQueryDB, job_id: CMJobID, label):
    possible = ["none", "id", "function"]
    if not label in possible:
        msg = f"Invalid label method {label!r} not in {possible!r}."
        raise ValueError(msg)
    if label == "none":
        return ""
    if label == "id":
        return job_id
    if label == "function":
        job = cq.get_job(job_id)
        return f"{job.command_desc}()"
    assert False
    #
    #
    #


def create_graph1(
    cq: CacheQueryDB, job_list: Collection[CMJobID], label, color: bool, processing: Collection[CMJobID]
):
    from mcdp_report.my_gvgen import GvGen

    print("Creating graph")
    job_list = list(job_list)
    print(f"create_graph1({job_list})")
    ggraph = GvGen()

    job2node = {}
    for job_id in job_list:
        job_label = get_node_label(cq, job_id, label)
        job2node[job_id] = ggraph.newItem(job_label)

        if color:
            ggraph.styleAppend(job_id, "style", "filled")
            ggraph.styleAppend(job_id, "fillcolor", get_color_for(job_id, cq, processing)["color"])
            ggraph.styleApply(job_id, job2node[job_id])
        else:
            ggraph.styleAppend(job_id, "style", "filled")
            ggraph.styleAppend(job_id, "fillcolor", "#c0c0c0")
        ggraph.styleAppend(job_id, "shape", "box")
        ggraph.styleApply(job_id, job2node[job_id])

    for job_id in job_list:
        # c = get_computation(job_id)
        # children_id = [x.job_id for x in c.depends]
        for child in cq.direct_children(job_id):
            # arrows follows flux of data
            print(f"{job2node[child]}->{job2node[job_id]}")
            ggraph.newLink(job2node[child], job2node[job_id])

    return ggraph


def create_graph2_clusters(
    cq, job_list: Collection[CMJobID], label, color: bool, processing: Collection[CMJobID]
):
    from mcdp_report.my_gvgen import GvGen

    print("Creating graph")

    ggraph = GvGen()

    cluster2jobs = defaultdict(lambda: set())
    job2cluster = {}
    for job_id in job_list:
        job = cq.get_job(job_id)
        cluster = job.defined_by[-1]
        cluster2jobs[cluster].add(job_id)
        job2cluster[job_id] = cluster

    job2node = {}

    cluster2node = {}

    rel_generated_color = "brown"

    for cluster, cluster_jobs in cluster2jobs.items():
        cluster_label = ""  # if compact else cluster
        if cluster == "root":
            cluster2node[cluster] = None
        else:
            cluster2node[cluster] = ggraph.newItem(cluster_label)
            ggraph.styleAppend("cluster", "style", "dashed")
            ggraph.styleAppend("cluster", "color", rel_generated_color)
            ggraph.styleApply("cluster", cluster2node[cluster])

        for job_id in cluster_jobs:
            # job = cq.get_job(job_id)

            job_label = get_node_label(cq, job_id, label)
            job2node[job_id] = ggraph.newItem(job_label, cluster2node[cluster])

            if color:
                ggraph.styleAppend(job_id, "style", "filled")
                ggraph.styleAppend(job_id, "fillcolor", get_color_for(job_id, cq, processing))
                ggraph.styleApply(job_id, job2node[job_id])
            else:
                ggraph.styleAppend(job_id, "style", "filled")
                ggraph.styleAppend(job_id, "fillcolor", "#c0c0c0")
            ggraph.styleAppend(job_id, "shape", "box")
            ggraph.styleAppend(job_id, "fontname", "Anka/Coder")

            ggraph.styleApply(job_id, job2node[job_id])

    # dependency
    for job_id in job_list:
        # c = get_computation(job_id)
        # children_id = [x.job_id for x in c.depends]
        for child in cq.direct_children(job_id):
            ggraph.newLink(job2node[child], job2node[job_id])

    # generation
    for cluster in cluster2jobs:
        if cluster != "root":
            link = ggraph.newLink(job2node[cluster], cluster2node[cluster])

            ggraph.propertyAppend(link, "color", rel_generated_color)
            ggraph.propertyAppend(link, "style", "dashed")

    return ggraph



from compmake.constants import DefaultsToConfig
from compmake.jobs.queries import top_targets
from compmake.plugins.backend_sge.sge_manager import SGEManager
from compmake.ui.commands import raise_error_if_manager_failed
from compmake.ui.helpers import ACTIONS, ui_command

__all__ = [
    "sgemake",
]


@ui_command(section=ACTIONS, dbchange=True)
def sgemake(
    job_list,
    context,
    cq,
    n: int = DefaultsToConfig("max_parallel_jobs"),
    recurse: bool = DefaultsToConfig("recurse"),
):
    """ Cluster equivalent of "make" using the Sun Grid Engine. """
    job_list = [x for x in job_list]

    if not job_list:
        db = context.get_compmake_db()
        job_list = list(top_targets(db=db))

    manager = SGEManager(context=context, cq=cq, recurse=recurse, num_processes=n)
    manager.add_targets(job_list)
    manager.process()
    return raise_error_if_manager_failed(manager)



from .compmake_test import CompmakeTest
from compmake import Context
from compmake.jobs import direct_children, get_job, jobs_defined
from compmake.storage.filesystem import StorageFilesystem
from compmake.ui.ui import clean_other_jobs
from nose.tools import istest
from compmake.structures import Cache
from compmake.jobs.manager import check_job_cache_state
from compmake.exceptions import CompmakeBug


def g2():
    print("returning g2")
    return "fd-gd-g2"


def gd(context):
    context.comp(g2)


def fd(context):
    return context.comp_dynamic(gd)


def i2():
    return "hd-id-i2"


def id(context):  # @ReservedAssignment
    context.comp(i2)


def hd(context):
    return context.comp_dynamic(id)


def summary(res):
    pass


def mockup6(context, both):
    res = []
    res.append(context.comp_dynamic(fd))
    if both:
        res.append(context.comp_dynamic(hd))
    context.comp(summary, res)


@istest
class TestDynamic6(CompmakeTest):
    def test_dynamic6(self):

        # first define with job and run
        mockup6(self.cc, both=True)
        db = self.db

        self.assertRaises(CompmakeBug, jobs_defined, job_id="hd", db=db)

        self.assert_cmd_success("make recurse=1")
        check_job_cache_state(job_id="hd", states=[Cache.DONE], db=db)
        self.assertEqual(jobs_defined(job_id="hd", db=db), set(["hd-id"]))

        # self.assert_cmd_success('graph compact=0 color=0 '
        #                         'cluster=1 filter=dot')

        self.assertJobsEqual("all", ["fd", "fd-gd", "fd-gd-g2", "hd", "hd-id", "hd-id-i2", "summary"])
        self.assertJobsEqual("done", ["fd", "fd-gd", "fd-gd-g2", "hd", "hd-id", "hd-id-i2", "summary"])

        # now redo it
        self.db = StorageFilesystem(self.root, compress=True)
        self.cc = Context(db=self.db)

        print("running again with both=False")
        mockup6(self.cc, both=False)
        clean_other_jobs(context=self.cc)

        self.assertJobsEqual("all", ["fd", "fd-gd", "fd-gd-g2", "summary"])

        job = get_job("summary", self.db)
        print("job.children: %s" % job.children)
        print("job.dynamic_children: %s" % job.dynamic_children)
        self.assertEqual(job.dynamic_children, {"fd": set(["fd-gd"])})
        self.assertEqualSet(direct_children("summary", self.db), ["fd", "fd-gd"])
        self.assert_cmd_success("ls")

        self.assert_cmd_success("make recurse=1")
        self.assertJobsEqual("all", ["fd", "fd-gd", "fd-gd-g2", "summary"])
        self.assertJobsEqual("done", ["fd", "fd-gd", "fd-gd-g2", "summary"])



""" The actual interface of some commands in commands.py """
from collections import defaultdict

from typing import Dict

from compmake.constants import CompmakeConstants

from ..jobs import get_job, get_job_cache, parse_job_list
from ..structures import Cache
from ..ui import VISUALIZATION, compmake_colored, ui_command
from ..utils import pad_to_screen


@ui_command(section=VISUALIZATION)
def stats(args, context, cq):
    """ Displays a coarse summary of the jobs state. """
    if not args:
        job_list = cq.all_jobs()
    else:
        job_list = parse_job_list(args, context=context, cq=cq)

    job_list = list(job_list)
    CompmakeConstants.aliases["last"] = job_list
    display_stats(job_list, context)


def display_stats(job_list, context):
    db = context.get_compmake_db()
    states_order = [
        Cache.NOT_STARTED,
        # Cache.IN_PROGRESS,
        Cache.FAILED,
        Cache.BLOCKED,
        Cache.DONE,
    ]
    # initialize counters to 0
    states2count = dict(list(map(lambda x: (x, 0), states_order)))

    function2state2count: Dict[str, Dict[str, int]] = {}
    total = 0

    for job_id in job_list:

        cache = get_job_cache(job_id, db=db)
        states2count[cache.state] += 1
        total += 1

        function_id = get_job(job_id, db=db).command_desc
        # initialize record if not present
        if not function_id in function2state2count:
            function2state2count[function_id] = dict(list(map(lambda x: (x, 0), states_order)) + [("all", 0)])
        # update
        function2state2count[function_id][cache.state] += 1
        function2state2count[function_id]["all"] += 1

        if total == 100:  # XXX: use standard method
            print("Loading a large number of jobs...\r")

    if total == 0:
        print(pad_to_screen("No jobs found."))
        return

        # print("Found %s jobs in total." % total)
    #
    #     for state in states_order:
    #         desc = "%30s" % Cache.state2desc[state]
    #         # colorize output
    #         desc = compmake_colored(desc, **state2color[state])
    #
    #         num = states2count[state]
    #         if num > 0:
    #             print("%s: %5d" % (desc, num))

    print("Summary by function name:")

    flen = max((len(x) + len("()")) for x in function2state2count)
    flen = max(flen, len("total"))
    states = [
        (Cache.DONE, "done"),
        (Cache.FAILED, "failed"),
        (Cache.BLOCKED, "blocked"),
        #         (Cache.IN_PROGRESS, 'in progress'),
        (Cache.NOT_STARTED, "to do"),
    ]

    totals = defaultdict(lambda: 0)
    for function_id in sorted(function2state2count):
        function_stats = function2state2count[function_id]
        alls = []
        for state, desc in states:
            num = function_stats[state]
            s = f"{num:5d} {desc}"
            if num > 0:
                s = compmake_colored(s, **Cache.state2color[state])
            alls.append(s)
            totals[state] += num
        s = ",".join(alls)
        function_id_pad = (function_id + "()").ljust(flen)
        print(f"    {function_id_pad}: {s}.")

    final = []
    for state, desc in states:
        s = f"{totals[state]:5d} {desc}"
        if totals[state] > 0:
            s = compmake_colored(s, **Cache.state2color[state])
        final.append(s)
    final = ",".join(final)
    print(f"    {'total'.rjust(flen)}: {final}.")



from .compmake_test import CompmakeTest
from nose.tools import istest
from compmake.exceptions import CompmakeDBError


def g2():
    return "g2"


def gd(context):
    context.comp(g2)


def fd(context):
    return context.comp_dynamic(gd)


def mockup7(context):
    context.comp_dynamic(fd)


@istest
class TestDynamic7(CompmakeTest):
    def test_dynamic7(self):
        # first define with job and run
        mockup7(self.cc)
        self.assert_cmd_success("make recurse=1; ls")

        # check that g2 is up to date
        self.assertEqual(self.up_to_date("fd-gd-g2"), True)

        # now clean its parent
        self.assert_cmd_success("clean fd")

        # job does not exist anynmore
        self.assertRaises(CompmakeDBError, self.up_to_date, "fd-gd-g2")

    def test_dynamic7_invalidate(self):
        # first define with job and run
        mockup7(self.cc)
        self.assert_cmd_success("make recurse=1; ls")

        # check that g2 is up to date
        self.assertEqual(self.up_to_date("fd-gd-g2"), True)

        # now invalidate the parent
        self.assert_cmd_success("invalidate fd")

        # job exists but not up to date
        self.assertEqual(self.up_to_date("fd-gd-g2"), False)



from nose.tools import istest

from .compmake_test import CompmakeTest


def f(a):
    print("f()")
    if not a == 2:
        raise ValueError("Expected 2, not %s" % a)


def g(context):
    """ This function returns a Promise to another job. """
    print("g()")
    return context.comp(g_delegated)


def g_delegated():
    print("g_delegated()")
    return 1 + 1


@istest
class TestDelegation(CompmakeTest):
    def test_delegation_1(self):
        context = self.cc

        g_res = context.comp_dynamic(g)
        context.comp(f, g_res)

        self.assert_cmd_success("ls")
        self.assert_cmd_success("stats")
        self.assert_cmd_success("make g()")
        self.assert_cmd_success("ls")
        self.assert_cmd_success("make g_delegated()")

    def test_delegation_2(self):
        context = self.cc

        g_res = context.comp_dynamic(g)
        context.comp(f, g_res)

        self.assert_cmd_success("ls")
        self.assert_cmd_success("stats")
        self.assert_cmd_success("make")



from .capture import *
from .colored import *
from .coloredterm import *
from .docstring import *
from .format_exceptions import *
from .frozen import *
from .instantiate_utils import *
from .memoize_imp import *
from .pickling_utils import *
from .proctitle import *
from .safe_pickle import *
from .strings_with_escapes import *
from .system_stats import *
from .system_stats import *
from .terminal_size import *
from .time_track import *
from .timedate import *
from .values_interpretation import *
from .which_imp import *
from .wildcards import *

# def find_print_statements():
#     import sys
#
#     class TracePrints(object):
#         def __init__(self):
#
#             self.stdout = sys.stdout
#
#         def write(self, s):
#             self.stdout.write("Writing %r\n" % s)
#             import traceback
#             traceback.print_stack(file=self.stdout)
#
#     sys.stdout = TracePrints()



import itertools
import math
import sys
import time
from collections import namedtuple
from datetime import datetime
from typing import Dict

from compmake import CompmakeGlobalState
from zuper_commons.text import indent
from .tracker import Tracker
from ..events import publish, register_handler
from ..state import get_compmake_config
from ..ui import compmake_colored
from ..ui.visualization import ui_error
from ..utils import get_length_on_screen, getTerminalSize, pad_to_screen_length

stream = sys.stderr

tracker = Tracker()


def system_status():
    stats = CompmakeGlobalState.system_stats
    if not stats.available():  # psutil not installed
        # TODO: use os.load
        return ""

    cpu = stats.avg_cpu_percent()
    cur_mem = stats.cur_phymem_usage_percent()
    swap = stats.cur_virtmem_usage_percent()

    console_status_style = get_compmake_config("console_status_style")
    if console_status_style == "normal":

        s_mem = f"mem {cur_mem:2.0f}%"
        if swap > 20:
            s_mem += f" swap {swap:2.0f}%"

        return f"cpu {cpu:2.0f}% {s_mem}"

    else:

        return f"cpu {make_bar(cpu, 7)}"


def make_bar(percent: float, n: int) -> str:
    on = "▉"
    off = "▁"

    ns = int(math.ceil(n * percent / 100))
    ons = on * ns
    offs = off * (n - ns)
    return f"[{ons}{offs}]"


def get_spins():
    # toutf = lambda x: [_.encode('utf8') for _ in x]
    toutf = lambda x: [_ for _ in x]

    # from_sequence = lambda x: toutf(_ for _ in x)

    #     spins = toutf(_ for _ in u"▉▊▋▌▍▎▏▎▍▌▋▊▉")

    def get_spin_fish(n):
        fish_right = ">))'>"
        fish_left = "<'((<"
        s = []
        for i in range(n):
            s.append(" " * i + fish_right)
        for i in range(n):
            s.append(" " * (n - i) + fish_left)
        m = max(len(_) for _ in s)
        # return [_.ljust(m).encode('utf8') for _ in s]

        return [_.ljust(m) for _ in s]

    options = []
    options.append(get_spin_fish(12))
    options.append(list("⣾⣽⣻⢿⡿⣟⣯⣷"))
    options.append(list("◐◓◑◒"))
    options.append(list("◰◳◲◱"))
    options.append(list("◴◷◶◵"))
    #     options.append(from_sequence(u"🕐🕑🕒🕓🕔🕕🕖🕗🕘🕙🕚🕛"))
    options.append(list("▙▛▜▟"))
    #     options.append(['-', '/', '|', '\\'])

    today = datetime.today()
    # change every 3 days
    i = today.day / 3
    i = int(math.ceil(i))
    res = options[i % len(options)]

    return res


spins = get_spins()


def spinner():
    spin_interval = get_compmake_config("console_status_delta") * 0.8
    t = time.time()
    i = t / spin_interval
    i = int(i) % len(spins)
    return spins[i]


def job_counts():
    ss = []

    console_status_style = get_compmake_config("console_status_style")

    styles = {
        "normal": {
            "done": dict(color="green", text="NUM done"),
            "failed": dict(color="red", text="NUM failed"),
            "blocked": dict(text="NUM blocked"),
            "ready": dict(color="yellow", text="NUM ready"),
            "processing": dict(color="blue", text="NUM proc"),
            "todo": dict(color="cyan", text="NUM todo"),
        },
        "compact": {
            "done": dict(color="green", text="NUM ✔"),
            "failed": dict(color="red", text="NUM ✗"),
            "blocked": dict(text="NUM ⌘"),
            "ready": dict(color="yellow", text="NUM ▴‍"),
            "processing": dict(color="blue", text="NUM ⚙"),
            "todo": dict(color="cyan", text="NUM ☯"),
        },
    }
    style = styles[console_status_style]

    values: Dict[str, int] = {
        "done": len(tracker.done),
        "processing": len(tracker.processing),
        "failed": len(tracker.failed),
        "blocked": len(tracker.blocked),
        "ready": len(tracker.ready),
        "todo": len(tracker.todo),
    }

    for k, v in values.items():
        if k not in style:
            continue
        if v > 0:
            sk = style[k]
            text = style[k]["text"].replace("NUM", str(v))
            if "color" in sk:
                text = compmake_colored(text, color=sk["color"])
            ss.append(text)

    return " ".join(ss)


def wait_reasons():
    if tracker.wait_reasons:
        s = "(wait: " + ",".join(tracker.wait_reasons.values()) + ")"
    else:
        s = ""
    return s


def current_slot(intervals):
    period = sum(intervals)
    index = int(time.time()) % period
    csum = 0
    for i, delta in enumerate(intervals):
        if (index >= csum) and (index < csum + delta):
            return i
        csum += delta
    return len(intervals) - 1


def display_rotating(strings, intervals, align_right=False):
    """ Rotates the display of the given strings.
        For now, we assume intervals to be round integers.
    """
    which = current_slot(intervals)
    L = max(get_length_on_screen(x) for x in strings)
    aligned = pad_to_screen_length(strings[which], L, align_right=align_right)
    return aligned


def get_string(level):
    if level == -3:
        return ""
    if level == -2:
        return "..."
    if level == -1:
        return "  %d proc." % len(tracker.status)
    X = []

    for job_id, status in tracker.status.items():
        x = []
        if level >= 1:
            x += [job_id]

        if level <= 1 or not job_id in tracker.status_plus:
            x += [status]
        elif job_id in tracker.status_plus:
            x += []
            stack = tracker.status_plus[job_id]
            for i, frame in enumerate(stack):
                if level >= 3:
                    x += [frame.name]

                if level >= 2:
                    # XXX: this is never used somehow, see tracker
                    # that's where the code is executed to display iterations
                    if isinstance(frame.iterations[0], int) and isinstance(frame.iterations[1], int):
                        x += ["%s of %s" % (frame.iterations[0] + 1, frame.iterations[1])]
                    else:
                        perc = frame.iterations[0] * 100.0 / frame.iterations[1]
                        x += ["%.1f%%" % perc]

                if level >= 4 and frame.iteration_desc is not None:
                    x += ["(%s)" % frame.iteration_desc]

                if i < len(stack) - 1:
                    x += [">>"]
        X += [" ".join(x)]
    return " ".join(X)


class Tmp:
    last_manager_loop = time.time()


def its_time():
    delta = float(get_compmake_config("console_status_delta"))
    t = time.time()
    dt = t - Tmp.last_manager_loop
    if dt > delta:
        Tmp.last_manager_loop = t
        return True
    else:
        return False


def handle_event_period(context, event):
    if its_time():
        handle_event(context, event)


ShowOption = namedtuple("Option", "length left right weight")


def handle_event(context, event):
    if not get_compmake_config("status_line_enabled"):
        return

    status = system_status()

    options_right = []

    if status:
        options_right.append("%s %s " % (status, job_counts()))
        options_right.append("%s %s %s" % (wait_reasons(), status, job_counts()))

    options_right.append(job_counts())

    sp = spinner()
    options_left = [sp]

    for level in [4, 3, 2, 1, 0, -1, -2, -3]:
        options_left.append(f" compmake {sp} {get_string(level)}")
    #         options_left.append(sp + '  ' + get_string(level))

    cols, _ = getTerminalSize()

    # Make all options together
    options = []
    for l, r in itertools.product(options_left, options_right):
        length = get_length_on_screen(l) + get_length_on_screen(r)
        weight = length
        options.append(ShowOption(length=length, weight=weight, right=r, left=l))

    # sort by length decreasing
    options.sort(key=lambda _: _.length)
    choice = None
    for _ in options:
        if _.length < cols:
            choice = _

    if choice is None:
        # cannot find anything?
        choice = options[0]

    nspaces = cols - get_length_on_screen(choice.right) - get_length_on_screen(choice.left)
    line = choice.left + " " * nspaces + choice.right

    # if get_compmake_config("console_status"):
    #     # if six.PY2:
    #     #     if isinstance(line, unicode):
    #     #         line = line.encode('utf-8')
    #     stream.write(line)
    #
    #     interactive = get_compmake_config("interactive")
    #     if interactive:
    #         stream.write("\r")
    #     else:
    #         stream.write("\n")

    publish(context, "ui-status-summary", string=line)


def manager_host_failed(context, event):
    s = "Host failed for job %s: %s" % (event.job_id, event.reason)
    s += indent(event.bt.strip(), "| ")
    ui_error(context, s)


if get_compmake_config("status_line_enabled"):
    register_handler("manager-loop", handle_event_period)
    register_handler("manager-progress", handle_event_period)
    register_handler("job-progress", handle_event)
    register_handler("job-progress-plus", handle_event)
    register_handler("job-stdout", handle_event)
    register_handler("job-stderr", handle_event)
    register_handler("manager-host-failed", manager_host_failed)



from nose.tools import istest

from compmake.utils import wildcard_to_regexp

from . import CompmakeTest


@istest
class Simple(CompmakeTest):
    def mySetUp(self):
        pass

    def testExists1(self):
        key = "not-existent"
        assert not key in self.db

    def testExists2(self):
        k = "ciao"
        v = {"complex": 123}
        db = self.db
        if k in db:
            del db[k]
        self.assertFalse(k in db)
        db[k] = v
        self.assertTrue(k in db)
        del db[k]
        self.assertFalse(k in db)
        db[k] = v
        del db[k]
        self.assertFalse(k in db)

    def testSearch(self):
        db = self.db

        def search(pattern):
            r = wildcard_to_regexp(pattern)
            for k in db.keys():
                if r.match(k):
                    yield k

        self.assertEqual([], list(search("*")))
        db["key1"] = 1
        db["key2"] = 1
        self.assertEqual([], list(search("ciao*")))
        self.assertEqual(["key1"], list(search("key1")))
        self.assertEqual(["key1"], list(search("*1")))
        self.assertEqual([], list(search("d*1")))



from compmake.unittests.compmake_test import CompmakeTest
from nose.tools import istest


def g2():
    pass


def g(context):
    context.comp(g2)


def h2():
    pass


def h(context):
    context.comp(h2)


def fd(context):
    context.comp_dynamic(g)

    if TestDynamicFailure.do_fail is not None:
        raise TestDynamicFailure.do_fail()

    context.comp_dynamic(h)


def mockup8(context):
    context.comp_dynamic(fd)


@istest
class TestDynamicFailure(CompmakeTest):
    do_fail = False

    def test_dynamic_failure1(self):
        mockup8(self.cc)
        # run it
        TestDynamicFailure.do_fail = ValueError
        self.assert_cmd_fail("make recurse=1")
        # we have three jobs defined
        self.assertJobsEqual("all", ["fd"])

    def test_dynamic_failure2(self):
        mockup8(self.cc)
        # run it
        TestDynamicFailure.do_fail = None
        self.assert_cmd_success("make recurse=1")
        # we have three jobs defined
        self.assertJobsEqual("all", ["fd", "fd-h", "fd-h-h2", "fd-g", "fd-g-g2"])
        self.assertJobsEqual("done", ["fd", "fd-h", "fd-h-h2", "fd-g", "fd-g-g2"])

        TestDynamicFailure.do_fail = ValueError
        self.assert_cmd_success("invalidate fd")
        self.assert_cmd_success("stats")
        self.assert_cmd_fail("make")
        self.assertJobsEqual("all", ["fd"])

    def test_dynamic_failure3(self):
        mockup8(self.cc)
        # run it
        TestDynamicFailure.do_fail = KeyboardInterrupt
        self.assert_cmd_fail("make recurse=1")
        # we have three jobs defined
        self.assertJobsEqual("all", ["fd"])



import sys

from compmake.context import Context


def f(*args):
    print("to-std-out")
    sys.stderr.write("to-std-err")
    return


def fails(*args):
    raise Exception("this function fails")


def mockup1(context: Context):
    comp = context.comp
    return comp(f, comp(f), comp(f, comp(f)))


def mockup2(context: Context):
    comp = context.comp

    comp(f, job_id="f1")
    comp(f, job_id="f2")
    res = comp(fails, job_id="fail1")
    comp(f, res, job_id="blocked")

    r5 = comp(f, job_id="f5")
    comp(f, r5, job_id="needs_redoing")

    comp(f, job_id="verylong" + "a" * 40)

    context.batch_command("make")
    context.batch_command("clean f2")
    context.batch_command("clean f5")


def mockup2_nofail(context: Context):
    comp = context.comp

    comp(f, job_id="f1")
    comp(f, job_id="f2")

    r5 = comp(f, job_id="f5")
    comp(f, r5, job_id="needs_redoing")

    comp(f, job_id="verylong" + "a" * 40)

    context.batch_command("make")
    context.batch_command("clean f2")
    context.batch_command("clean f5")


def mockup_recursive_5(context):
    recursive(context, 5)


def recursive(context, v):
    if v == 0:
        print("finally!")
        return

    context.comp_dynamic(recursive, v - 1, job_id="r%d" % v)



from typing import Callable
from .. import set_compmake_status, CompmakeConstants
from .compmake_test import CompmakeTest
from nose.tools import istest


def f1() -> Callable:
    print("done")
    return lambda _: None


@istest
class TestUnpickable(CompmakeTest):
    def mySetUp(self) -> None:
        # TODO: use tmp dir
        set_compmake_status(CompmakeConstants.compmake_status_embedded)

    def test_unpickable_result(self) -> None:
        self.comp(f1)
        self.cc.batch_command("clean")

        self.assert_cmd_fail("make")
        # since dill implemented
        # self.assert_cmd_success('make')



import sys
from typing import List

from compmake.structures import Cache
from six import StringIO

from .. import get_compmake_config
from ..events import register_handler
from ..ui import compmake_colored, ui_message
from ..utils import get_length_on_screen, get_screen_columns, pad_to_screen, pad_to_screen_length

# sys.stdout will be changed later
stream = sys.stdout


class Storage(object):
    max_len = 0
    last_job_id = None


def plot_with_prefix(job_id, lines, is_stderr):
    for line in lines:

        formats = "%%%ds" % Storage.max_len

        prefix = formats % job_id
        prefix_empty = formats % ""

        if Storage.last_job_id != job_id:
            Storage.last_job_id = job_id
            # prefix = colored(prefix, color='cyan', attrs=['dark'])
            prefix = compmake_colored(prefix, color="cyan")
        else:
            prefix = prefix_empty

        if is_stderr:
            sep = compmake_colored("|", "red")
        else:
            sep = compmake_colored("|", "cyan")
        write_screen_line(line)

        # Now let's take lines that do not fit the length

        # This has problems with escape characters
        # (in addition to get_screen_columns() not functioning sometime.)

        # split_lines = False
        # if split_lines:
        #     max_space = (get_screen_columns() - 1
        #                  - get_length_on_screen('%s%s%s' % (prefix, sep, '')))
        #
        #     sublines = clip_to_length(line, max_space)
        #
        #     for a, subline in enumerate(sublines):
        #         if a == 0:
        #             screen_line = '%s%s%s' % (prefix, sep, subline)
        #         else:
        #             screen_line = '%s%s%s' % (prefix_empty, ' ', subline)
        #
        #         screen_line = pad_to_screen(screen_line)
        #         write_line_endl(screen_line)
        #
        # else:
        #     pad = True
        #     if pad:
        #         write_screen_line(line)
        #     else:
        #         write_line_endl(line)


def write_line_endl_w(x: str, ss):
    xl = x + "\n"
    write_on_buffer(xl, ss)


def write_on_buffer(xl, ss):
    if isinstance(ss, StringIO):
        ss.write(xl)
    else:
        if hasattr(ss, "buffer"):
            # noinspection PyUnresolvedReferences
            ss.buffer.write(xl.encode("utf-8"))
        else:
            ss.write(xl)
    ss.flush()


def write_line(x: str):
    write_on_buffer(x, stream)


def write_line_endl(x: str):
    write_line_endl_w(x, stream)


def write_screen_line(s: str):
    """ Writes and pads """
    # TODO: check that it is not too long
    s = pad_to_screen(s)
    write_line_endl(s)


def plot_normally(job_id, lines, is_stderr):
    for line in lines:
        if Storage.last_job_id != job_id:
            Storage.last_job_id = job_id
            # job_name = colored('%s' % job_id, color='cyan')
            marker = "*" if is_stderr else ""
            header = pad_to_screen(f"___ {job_id} {marker}", pad="_")
            header = compmake_colored(header, color="cyan")
            write_screen_line(header)

        max_size = get_screen_columns()
        # if debug_padding:
        # prefix = compmake_colored('>', color='red')
        # postfix = compmake_colored('<', color='blue')
        # else:
        prefix = ""
        postfix = ""

        # reproducing 3.5.6: safe
        # write_screen_line(line)

        # if True:  # need to check unicode anyway
        # 3.5.10 -- most recent

        sublines = break_lines(prefix, line, postfix, max_size)

        if sublines:
            for s in sublines:
                s = pad_to_screen(s)
                write_line(s)
            # write_line(sublines[-1]+'\n')
            write_line("\n")
        #
        # elif False:
        #     sublines = break_lines_and_pad(prefix, line, postfix, max_size)
        #     for s in sublines:
        #         write_screen_line(s)
        # else:
        #     # 3.5.6
        #     write_screen_line(line)


def break_lines(prefix: str, line: str, postfix: str, max_size: int):
    # Now let's take lines that do not fit the length
    prefix_len = get_length_on_screen(prefix)
    postfix_len = get_length_on_screen(postfix)

    max_space = max_size - postfix_len - prefix_len

    if max_space < 10:
        msg = "Weird max space: %s" % max_space
        msg += " max_size: %s prefix: %s postfix: %s" % (max_size, prefix_len, postfix_len)
        raise ValueError(msg)

    # XXX: might have problems with colors
    sublines = clip_to_length(line, max_space)

    lines = []
    for _, subline in enumerate(sublines):
        # pad = '+' if debug_padding else ' '
        #         pad = ' '
        #         subline = pad_to_screen_length(subline, max_space, pad=pad)
        line = "%s%s%s" % (prefix, subline, postfix)
        lines.append(line)
    return lines


# @contract(prefix='str', line='str', postfix='str', returns='list[>=1]x(str)')
def break_lines_and_pad(prefix, line, postfix, max_size):
    # Now let's take lines that do not fit the length
    prefix_len = get_length_on_screen(prefix)
    postfix_len = get_length_on_screen(postfix)

    max_space = max_size - postfix_len - prefix_len

    # XXX: might have problems with colors
    sublines = clip_to_length(line, max_space)

    lines = []
    for _, subline in enumerate(sublines):
        # pad = '+' if debug_padding else ' '
        pad = " "
        subline = pad_to_screen_length(subline, max_space, pad=pad)
        line = "%s%s%s" % (prefix, subline, postfix)
        lines.append(line)
    return lines


def handle_event(event, is_stderr: bool):
    job_id = event.kwargs["job_id"]
    lines = event.kwargs["lines"]

    Storage.max_len = max(Storage.max_len, len(job_id))

    if Storage.max_len < 15:
        plot_with_prefix(job_id, lines, is_stderr)
    else:
        plot_normally(job_id, lines, is_stderr)


def clip_to_length(line: str, max_len: int) -> List[str]:
    if max_len <= 0:
        msg = "Max length should be positive."
        raise ValueError(msg)
    sublines = []
    while len(line):
        if len(line) < max_len:
            sublines.append(line)
            break
        initial, rest = clip_up_to(line, max_len)
        # clip = min(len(line), max_len)
        # subline = line[:clip]
        sublines.append(initial)
        line = rest
    return sublines


def clip_up_to(line: str, max_len: int):
    if get_length_on_screen(line) <= max_len:
        return line, ""
    for i in reversed(range(len(line))):
        if get_length_on_screen(line[:i]) <= max_len:
            return line[:i], line[i:]
    return line[:max_len], line[max_len:]


def handle_event_stdout(event, context):
    if get_compmake_config("echo") and get_compmake_config("echo_stdout"):
        handle_event(event, False)


def handle_event_stderr(event, context):
    echo = get_compmake_config("echo")
    echo_stderr = get_compmake_config("echo_stderr")
    if echo and echo_stderr:
        handle_event(event, True)


register_handler("job-stdout", handle_event_stdout)
register_handler("job-stderr", handle_event_stderr)


def color_done(s):
    return compmake_colored(s, **Cache.styles[Cache.DONE])


def color_processing(s):
    return compmake_colored(s, **Cache.styles[Cache.PROCESSING])


def color_failed(s):
    return compmake_colored(s, **Cache.styles[Cache.FAILED])


def color_blocked(s):
    return compmake_colored(s, **Cache.styles[Cache.BLOCKED])


def color_ready(s):
    return compmake_colored(s, **Cache.styles["ready"])


def handle_job_done(event, context):
    job_id = event.kwargs["job_id"]
    desc = f"{Cache.state2desc[Cache.DONE]:>10}"
    glyph = Cache.glyphs[Cache.DONE]
    ui_message(context, color_done(f"{glyph} {desc} {job_id}"))


def handle_job_failed(event, context):
    job_id = event.kwargs["job_id"]
    desc = f"{Cache.state2desc[Cache.FAILED]:>10}"
    glyph = Cache.glyphs[Cache.FAILED]
    ui_message(context, color_failed(f"{glyph} {desc} {job_id}"))


def handle_job_processing(event, context):
    job_id = event.kwargs["job_id"]
    desc = f"{Cache.state2desc[Cache.PROCESSING]:>10}"
    glyph = Cache.glyphs[Cache.PROCESSING]
    ui_message(context, color_processing(f"{glyph} {desc} {job_id}"))


#  9 ✔ 1 ⚙ 2 ▴‍
def handle_job_blocked(event, context):
    job_id = event.kwargs["job_id"]
    blocking_job_id = event.kwargs["blocking_job_id"]
    desc = f"{Cache.state2desc[Cache.BLOCKED]:>10}"
    glyph = Cache.glyphs[Cache.BLOCKED]
    ui_message(context, color_blocked(f"{glyph} {desc} {job_id}") + f" because of {blocking_job_id}")


def handle_job_ready(event, context):
    job_id = event.kwargs["job_id"]
    glyph = Cache.glyphs["ready"]
    desc = f'{"ready":>10}'
    ui_message(context, color_ready(f"{glyph} {desc} {job_id}"))


register_handler("manager-job-done", handle_job_done)
register_handler("manager-job-failed", handle_job_failed)
register_handler("manager-job-processing", handle_job_processing)
register_handler("manager-job-ready", handle_job_ready)
register_handler("manager-job-blocked", handle_job_blocked)



from multiprocessing import active_children
from nose.tools import istest
from compmake.unittests.compmake_test import CompmakeTest


def g(b1, b2):
    pass


def f(context, level):
    if level == 0:
        context.comp(g, 1, 1)
    else:
        context.comp_dynamic(f, level - 1)
        # if level >= 2 or random.random() < 0.5:
        context.comp_dynamic(f, level - 1)


def mockup(context):
    context.comp_dynamic(f, 5)


@istest
class TestDynamic9(CompmakeTest):

    if False:

        def test_dynamic9_redefinition(self):
            mockup(self.cc)
            self.assert_cmd_success("make recurse=1")

            self.assertEqual(len(self.get_jobs("g()")), 32)
            self.assertEqual(len(self.get_jobs("f()")), 63)

            self.assert_cmd_success("clean")
            self.assertJobsEqual("all", ["f"])

            self.assert_cmd_success("make recurse=1")

            self.assertEqual(len(self.get_jobs("g()")), 32)
            self.assertEqual(len(self.get_jobs("f()")), 63)

    def test_dynamic9_redefinition2(self):
        mockup(self.cc)
        self.assert_cmd_success("parmake recurse=1")
        # ac =  active_children()
        # print('active children: %s' % ac)
        # showtree()
        # for a in ac:
        #     Process
        assert not active_children()
        self.assertEqual(len(self.get_jobs("g()")), 32)
        self.assertEqual(len(self.get_jobs("f()")), 63)

        self.assert_cmd_success("clean")
        self.assertJobsEqual("all", ["f"])

        self.assert_cmd_success("parmake recurse=1")
        assert not active_children()

        self.assertEqual(len(self.get_jobs("g()")), 32)
        self.assertEqual(len(self.get_jobs("f()")), 63)


# import os
# def showtree():
#     print('showing process tree')
#     parent = psutil.Process(os.getpid())
#     for child in parent.children(recursive=True):
#         print("child: %s"%child)
#         child.kill()

#     if including_parent:
#         parent.kill()
#
# ## get the pid of this program
# pid=os.getpid()
#
# ## when you want to kill everything, including this program
# killtree(pid)



from compmake.jobs.priority import compute_priorities
from compmake.ui.helpers import COMMANDS_ADVANCED, ui_command


@ui_command(section=COMMANDS_ADVANCED)
def debug_priority(non_empty_job_list, context, cq):
    """ Shows the priority of jobs. """
    jobs = list(non_empty_job_list)
    priorities = compute_priorities(all_targets=jobs, cq=cq)

    sorted_jobs = sorted(jobs, key=lambda x: priorities[x])
    for job_id in sorted_jobs:
        p = priorities[job_id]
        print("%5s %s" % (p, job_id))



# import base64
# import logging
# from multiprocessing import Pool
# import pickle
# import subprocess
# import sys
# import time
# import traceback
#
# from cjson import encode, decode, EncodeError, DecodeError  #
# @UnresolvedImport
# from compmake import CompmakeConstants
#
# from ..events import register_handler, remove_all_handlers, broadcast_event
# from ..jobs import (colorize_loglevel, get_job, set_job_userobject,
# set_job_cache,

# mark_as_failed)
# from ..structures import Cache, CompmakeException, JobFailed, HostFailed
# from ..ui import info, error
# from ..utils import OutputCapture, setproctitle
# from .cluster_conf import Host
# from .manager import Manager
# from .manager_local import FakeAsync
# from compmake.context import Context
# import warnings
#
#
# __all__ = ['ClusterManager']
#
#
# class ClusterManager(Manager):
#     def __init__(self, context, cq, hosts):
#         ''' Hosts: name -> Host '''
#         Manager.__init__(self, context=context, cq=cq)
#
#         self.hosts = hosts
#
#         # multiply hosts
#         newhosts = {}
#         for  hostconf in self.hosts.values():
#             for n in range(hostconf.processors):
#                 newname = hostconf.name + ':%s' % n
#                 h = hostconf._asdict()
#                 h['instance'] = n
#                 newhosts[newname] = Host(**h)
#
#         self.hosts = newhosts
#
#     def process_init(self):
#         self.failed_hosts = set()
#         self.hosts_processing = []
#         self.hosts_ready = self.hosts.keys()
#         # job-id -> host
#         self.processing2host = {}
#         self.pool = Pool(processes=len(self.hosts_ready))
#
#     def process_finished(self):
#         if self.failed_hosts:
#             error('The following hosts failed: %s.' %
#                   ", ".join(list(self.failed_hosts)))
#
#     def can_accept_job(self, reasons_why_not):
#         # only one job at a time
#         return self.hosts_ready
#
#     # XXX some confusion with names
#     def my_host_failed(self, host):
#         self.failed_hosts.add(host)
#         while host in self.hosts_ready:
#             self.hosts_ready.remove(host)
#         info('Host %s failed, removing from stack (failed now %s)' %
#                  (host, self.failed_hosts))
#
#     def job_failed(self, job_id):
#         Manager.job_failed(self, job_id)
#         self.release(job_id)
#
#     def host_failed(self, job_id):
#         Manager.host_failed(self, job_id)
#         host = self.processing2host[job_id]
#         self.my_host_failed(host)
#         self.release(job_id)
#
#     def job_succeeded(self, job_id):
#         Manager.job_succeeded(self, job_id)
#         self.release(job_id)
#
#     def release(self, job_id):
#         slave = self.processing2host[job_id]
#         del self.processing2host[job_id]
#         if not slave in self.failed_hosts:
#             self.hosts_ready.append(slave)
#             # info("Putting %s into the stack again (failed: %s)" %
#             #     (slave, self.failed_hosts))
#         else:
#             pass
#             # info("Not reusing host %s because it failed (failed: %s)" %
#             #    (slave, self.failed_hosts))
#
#     def instance_job(self, job_id):
#         slave = self.hosts_ready.pop()
#         self.processing2host[job_id] = slave
#
#         host_config = self.hosts[slave]
#
#         f = cluster_job
#         nice = None
#         fargs = self.context, job_id, host_config.name,
# host_config.username, nice
#
#         debug = False
#         if not debug:
#             async_result = self.pool.apply_async(f, fargs)
#         else:
#             # Useful for debugging the logic:
#             # run serially instead of in parallel
#             async_result = FakeAsync(f, *fargs)
#
#         return async_result
#
#     def event_check(self):
#         pass
#
#
# def compmake_slave():
#     s = StreamCon(sys.stdin, sys.stdout)
#
#     def msg(x):
#         sys.stderr.write('%s: %s' % ('slave', x))
#         sys.stderr.write('\n')
#         sys.stderr.flush()
#
#     try:
#         job_id = s.read()
#         # Note the order of first, second, third below.
#
#         # MUST BE first
#         remove_all_handlers()
#
#         # MUST BE second
#         warnings.warn('this must be changed')
#         context = Context(db={})
#         capture = OutputCapture(context=context, prefix=job_id,
#                                 echo_stdout=False, echo_stderr=False)
#         try:
#             # MUST BE third
#             actual = s.read()
#         except Exception as e:
#             msg = ('I could not deserialize the data or the function. '
#                    'Make sure that your package is on the python path. '
#                    'My python path is the following:\n%s'
#                     '\n\n\nA confusing message will appear next:\n\n%s' %
#                     (sys.path, traceback.format_exc(e)))
#             raise Exception(msg)
#
#         function, args, kwargs = actual
#
#         def handler(context, event):
#             s.write(('event', event))
#
#         register_handler("*", handler)  # third (otherwise stdout dirty)
#
#         # TODO: add whether we should just capture and not echo
#         old_emit = logging.StreamHandler.emit
#
#         def my_emit(_, log_record):
#             msg = colorize_loglevel(log_record.levelno, log_record.msg)
#             #  levelname = log_record.levelname
#             name = log_record.name
#             # print('%s:%s:%s' % (name, levelname, msg))
#             # TODO: use special log event?
#             sys.stderr.write('%s:%s\n' % (name, msg))
#
#         logging.StreamHandler.emit = my_emit
#
#         try:
#             result = function(*args, **kwargs)
#         except Exception as e:
#             s.write(('failure', (str(e), traceback.format_exc(e))))
#             return CompmakeConstants.RET_CODE_JOB_FAILED
#         finally:
#             capture.deactivate()
#             logging.StreamHandler.emit = old_emit
#
#         s.write(('success', result))
#
#     except Exception as e:
#         s.write(('host-failure', (str(e), traceback.format_exc(e))))
#         msg('Emergency exit -- something wrong happened')
#         sys.exit(1)
#
#     sys.exit(0)
#
#
# # TODO: what about wrong hostname?
# def cluster_job(context, job_id, hostname, username=None, nice=None):
#     setproctitle('%s %s' % (job_id, hostname))
#
#     if username:
#         connection_string = '%s@%s' % (username, hostname)
#     else:
#         connection_string = hostname
#
#     command = ['ssh', connection_string,
#                # '-X',
#                'compmake_slave']
#
#     try:
#         PIPE = subprocess.PIPE
#         p = subprocess.Popen(command, stdout=PIPE, stdin=PIPE, stderr=PIPE)
#
#         stream = StreamCon(p.stdout, p.stdin)
#         stream.write(job_id)
#         job = get_job(job_id)
#         stream.write(job.get_actual_command())
#         # todo: mark in progress?
#         while True:
#             what, result = stream.read()
#
#             if what == 'success' or what == 'failure':
#                 if p.stderr is not None:
#                     st = p.stderr.read()
#                     if st:
#                         print('Warning, dirty stderr for ssh: %r' % st)
#
#             if what == 'success':
#                 user_object = result
#                 set_job_userobject(job_id, user_object)
#                 cache = Cache(Cache.DONE)
#                 cache.state = Cache.DONE
#                 cache.timestamp = time.time()
#                 walltime = 1  # FIXME
#                 cputime = walltime  # FIXME
#                 cache.walltime_used = walltime
#                 cache.cputime_used = cputime
#                 cache.host = hostname
#                 set_job_cache(job_id, cache)
#                 return True
#
#             if what == 'failure':
#                 error, error_bt = result
#                 mark_as_failed(job_id, error, error_bt)
#                 raise JobFailed('Job %r failed' % job_id)
#
#             if what == 'host-failure':
#                 error, error_bt = result
#                 raise HostFailed(error_bt)
#
#             if what == 'event':
#                 event = result
#                 event.kwargs['remote'] = True
#                 broadcast_event(context, event)
#                 continue
#
#             raise Exception('Unknown what: %r' % what)
#
#     except ComException as e:
#         raise HostFailed('Communication with host %s failed (%s)' %
#                          (hostname, e))
#     except HostFailed:
#         raise
#     except JobFailed:
#         raise
#     except Exception as e:
#         # TODO: different exception
#         bt = traceback.format_exc(e)
#         raise CompmakeException('BUG: strange exception %s' % bt)
#
#
# class ComException(Exception):
#     ''' Communication exception. '''
#     pass
#
#
# class StreamCon(object):
#     ''' Simple communication stream.
#
#         It sends and receives python objects by enclosing them in a json
#         packet:
#
#             object -> pickled -> base64 -> json packet
#     '''
#
#     def __init__(self, stdin, stdout):
#         self.stdin = stdin
#         self.stdout = stdout
#
#     def write_json(self, dic):
#         try:
#             self.stdout.write(encode(dic))
#             self.stdout.write("\n")
#             self.stdout.flush()
#         except IOError, ex:
#             raise ComException("IOError while writing: %s" % ex)
#         except EncodeError, ex:
#             raise ComException("Cannot encode json. \n\t %s '''%s'''\n" %
#                                        (str(ex), dic))
#
#     def read_json(self):
#         try:
#             line = self.stdin.readline()
#             if not line:
#                 raise ComException("Broken communication")
#             return decode(line)
#         except IOError, ex:
#             raise ComException("IOError while reading: %s" % ex)
#         except DecodeError, ex:
#             raise ComException("Cannot decode json: '%s' \n\t %s\n" %
#                                (line, str(ex)))
#
#     def write(self, ob):
#         binary = pickle.dumps(ob)
#         b64 = base64.b64encode(binary)
#         self.write_json({'method': 'pickle',
#                          'base64': b64})
#
#     def read(self):
#         response = self.read_json()
#         if response['method'] != 'pickle':
#             raise Exception('Expected pickle, got %r' % response)
#         b64 = response['base64']
#         binary = base64.b64decode(b64)
#         data = pickle.loads(binary)
#         return data



import imp
import os
import pwd

from ..exceptions import UserError
from ..ui import COMMANDS_ADVANCED, ui_command

# noinspection PyShadowingBuiltins
from ..ui.visualization import ui_error, ui_info


@ui_command(section=COMMANDS_ADVANCED)
def reload(module, context):  # @ReservedAssignment
    """ Reloads a module.

        Usage::

            reload module=my_module

    """

    if module.startswith("compmake"):
        # noinspection PyBroadException
        try:
            dave = pwd.getpwuid(os.getuid())[0]
        except:
            dave = "Dave"
        ui_error(context, f"I'm sorry, {dave}. I'm afraid I can't do that.")
        return

    try:
        # otherwise import("A.B") returns A instead of A.B
        m = __import__(module, fromlist=["dummy"])
    except Exception as e:
        raise UserError(f'Cannot find module "{module}".') from e

    try:
        imp.reload(m)
    except Exception as e:
        msg = "Obtained this exception while reloading the module"
        raise UserError(msg) from e

    ui_info(context, f'Reloaded module "{module}".')



import time

__all__ = [
    "AvgSystemStats",
]

try:
    import psutil  # @UnusedImport
except ImportError:
    from compmake import logger

    logger.warning('Package "psutil" not found; load balancing ' "and system stats (CPU, MEM) not available.")


class AvgSystemStats(object):
    """ Collects average statistics about the system using psutil. """

    # noinspection PyUnresolvedReferences
    def __init__(self, interval, history_len):
        """

        :param interval: Collect statistics according to this interval.
        :param history_len: Use this many to compute avg/max statistics.
        """

        self.interval = interval
        self.history_len = history_len
        try:
            import psutil  # @UnresolvedImport @Reimport
        except:
            self._available = False
        else:
            self._available = True

            self.cpu = Collect("cpu", lambda: psutil.cpu_percent(interval=0), interval, history_len)

            try:
                # new in 0.8
                _ = psutil.virtual_memory().percent
                get_mem = lambda: psutil.virtual_memory().percent
            except:
                get_mem = lambda: psutil.phymem_usage().percent

            self.mem = Collect("mem", get_mem, interval, history_len)
            try:
                # new in 0.8
                _ = psutil.swap_memory().percent
                get_mem = lambda: psutil.swap_memory().percent
            except:
                get_mem = lambda: psutil.virtmem_usage().percent

            self.swap_mem = Collect("swap", get_mem, interval, history_len)

    def avg_cpu_percent(self):
        self._check_available()
        return self.cpu.get_avg()

    def max_cpu_percent(self):
        self._check_available()
        return self.cpu.get_max()

    def avg_phymem_usage_percent(self):
        self._check_available()
        return self.mem.get_avg()

    def cur_phymem_usage_percent(self):
        self._check_available()
        return self.mem.get_cur()

    def cur_virtmem_usage_percent(self):
        self._check_available()
        return self.swap_mem.get_cur()

    def available(self):
        """ returns false if psutil is not installed """
        return self._available

    def _check_available(self):
        if not self._available:
            msg = "Sorry, psutil not available."
            raise ValueError(msg)


class Collect(object):
    def __init__(self, name, function, interval, history_len):
        self.name = name
        self.function = function
        self.interval = interval
        self.history_len = history_len
        self.last_time = None
        self.values = []

    def get_cur(self):
        """ Returns the last value. """
        self.update_if_necessary()
        return self.values[-1]

    def get_min(self):
        self.update_if_necessary()
        return min(self.values)

    def get_max(self):
        self.update_if_necessary()
        return max(self.values)

    def get_avg(self):
        self.update_if_necessary()
        return sum(self.values) * 1.0 / len(self.values)

    def update_if_necessary(self):
        if self.values and self.time_from_last() < self.interval:
            return

        self.values.append(self.function())
        self.last_time = time.time()

        if len(self.values) > self.history_len:
            self.values.pop(0)
            # print('%s: %s' % (self.name, self.values))

    def time_from_last(self):
        if self.last_time is None:
            return self.interval * self.history_len * 2
        else:
            return time.time() - self.last_time



__all__ = [
    "interpret_strings_like",
]


def interpret_strings_like(args, reference_value):
    if not isinstance(args, list):
        args = [args]

    if isinstance(reference_value, str):
        value = " ".join(args)
    elif isinstance(reference_value, bool):
        if len(args) > 1:
            raise ValueError("Too many arguments for bool.")
        try:
            value = eval(args[0])
        except:
            raise ValueError("Could not parse %s " % args[0])
        value = bool(value)

    elif isinstance(reference_value, int):
        if len(args) > 1:
            raise ValueError("Too many arguments for int.")
        value = int(args[0])
    elif isinstance(reference_value, float):
        if len(args) > 1:
            raise ValueError("Too many arguments for float.")
        value = float(args[0])
    else:
        # XXX: security risk?
        try:
            value = eval(args[0])
        except:
            raise ValueError("Could not parse %s " % args[0])

    return value



from compmake.structures import CMJobID, IntervalTimer
from zuper_commons.types import check_isinstance
from .dependencies import collect_dependencies, substitute_dependencies
from .storage import get_job_args, job_userobject_exists
from ..exceptions import CompmakeBug
from ..structures import Job

__all__ = [
    "job_compute",
]


def get_cmd_args_kwargs(job_id: CMJobID, db):
    """ Substitutes dependencies and gets actual cmd, args, kwargs. """
    command, args, kwargs = get_job_args(job_id, db=db)
    kwargs = dict(**kwargs)
    # Let's check that all dependencies have been computed
    all_deps = collect_dependencies(args) | collect_dependencies(kwargs)
    for dep in all_deps:
        if not job_userobject_exists(dep, db):
            msg = f"Dependency {dep!r} was not done."
            raise CompmakeBug(msg)
    args2 = substitute_dependencies(args, db=db)
    kwargs2 = substitute_dependencies(kwargs, db=db)
    return command, args2, kwargs2


class JobCompute:
    # currently executing job id
    current_job_id = None


def job_compute(job: Job, context):
    """ Returns a dictionary with fields "user_object" and "new_jobs" """
    check_isinstance(job, Job)
    job_id = job.job_id
    db = context.get_compmake_db()

    int_load_results = IntervalTimer()

    command, args, kwargs = get_cmd_args_kwargs(job_id, db=db)

    int_load_results.stop()

    JobCompute.current_job_id = job_id
    if job.needs_context:
        args = tuple(list([context]) + list(args))

        int_compute = IntervalTimer()
        res = execute_with_context(
            db=db, context=context, job_id=job_id, command=command, args=args, kwargs=kwargs
        )
        int_compute.stop()

        assert isinstance(res, dict)
        assert len(res) == 2, list(res.keys())
        assert "user_object" in res
        assert "new_jobs" in res

        res["int_load_results"] = int_load_results
        res["int_compute"] = int_compute
        res["int_gc"] = IntervalTimer()
        return res
    else:
        int_compute = IntervalTimer()
        user_object = command(*args, **kwargs)
        int_compute.stop()

        res = dict(user_object=user_object, new_jobs=[])

        res["int_load_results"] = int_load_results
        res["int_compute"] = int_compute
        res["int_gc"] = IntervalTimer()

        return res


def execute_with_context(db, context, job_id, command, args, kwargs):
    """ Returns a dictionary with fields "user_object" and "new_jobs" """
    from compmake.context import Context

    assert isinstance(context, Context)
    from compmake.jobs.storage import get_job

    cur_job = get_job(job_id=job_id, db=db)
    context.currently_executing = cur_job.defined_by + [job_id]

    already = set(context.get_jobs_defined_in_this_session())
    context.reset_jobs_defined_in_this_session([])

    if args:
        if isinstance(args[0], Context) and args[0] != context:
            msg = "%s(%s, %s)" % (command, args, kwargs)
            raise ValueError(msg)

    # context is one of the arguments
    assert context in args

    res = command(*args, **kwargs)

    generated = set(context.get_jobs_defined_in_this_session())
    context.reset_jobs_defined_in_this_session(already)
    return dict(user_object=res, new_jobs=generated)


#    if generated:
#        if len(generated) < 4:
#            # info('Job %r generated %s.' % (job_id, generated))
#            pass
#        else:
#            # info('Job %r generated %d jobs such as %s.' %
#            # (job_id, len(generated), sorted(generated)[:M]))
#            pass
#            # # now remove the extra jobs that are not needed anymore

#     extra = []

# FIXME this is a RACE CONDITION -- needs to be done in the main thread
# from compmake.ui.visualization import info

# info('now cleaning up; generated = %s' % generated)
#
#     if False:
#         for g in all_jobs(db=db):
#             try:
#                 job = get_job(g, db=db)
#             except:
#                 continue
#             if job.defined_by[-1] == job_id:
#                 if not g in generated:
#                     extra.append(g)
#
#         for g in extra:
#             #info('Previously generated job %r (%s) removed.' % (g,
#             # job.defined_by))
#             delete_all_job_data(g, db=db)
#
#             #     from compmake.jobs.manager import
#             # clean_other_jobs_distributed
#             #     clean_other_jobs_distributed(db=db, job_id=job_id,
#             # new_jobs=generated)



import os
import stat
import traceback
from glob import glob
from os.path import basename
from typing import Iterator, List, NewType

import dill

from compmake import logger
from compmake.exceptions import CompmakeBug, SerializationError
from compmake.utils import safe_pickle_dump, safe_pickle_load
from zuper_commons.fs import (
    DirPath,
    FilePath,
    find_pickling_error,
    safe_read,
    safe_write,
    write_ustring_to_utf8_file,
)
from zuper_commons.types import ZException

__all__ = [
    "StorageFilesystem",
    "StorageKey",
]
if True:

    def track_time(x):
        return x


else:
    from ..utils import TimeTrack

    track_time = TimeTrack.decorator

trace_queries = False

StorageKey = NewType("StorageKey", str)


class StorageFilesystem:
    basepath: DirPath
    checked_existence: bool
    method: str
    """ `pickle` or `dill` """
    file_extension: str

    def __init__(self, basepath: DirPath, compress: bool = False):
        self.basepath = os.path.realpath(basepath)
        self.checked_existence = False
        self.method = method = "pickle"

        if compress:
            self.file_extension = f".{method}.gz"
            others = list(self.keys0(f".{method}"))
        else:
            self.file_extension = f".{method}"
            others = list(self.keys0(f".{method}.gz"))
        if others:
            msg = "Extension is %s but found %s files with other extension." % (
                self.file_extension,
                len(others),
            )
            raise ZException(msg)

        # create a bunch of files that contain shortcuts
        create_scripts(self.basepath)

    def __repr__(self) -> str:
        return f"FilesystemDB({self.basepath!r};{self.file_extension})"

    @track_time
    def sizeof(self, key: StorageKey) -> int:
        filename = self.filename_for_key(key)
        statinfo = os.stat(filename)
        return statinfo.st_size

    @track_time
    def __getitem__(self, key: StorageKey) -> object:
        if trace_queries:
            logger.debug("R %s" % str(key))

        self.check_existence()

        filename = self.filename_for_key(key)

        if not os.path.exists(filename):
            msg = f"Could not find key {key!r}."
            msg += f"\n file: {filename}"
            raise CompmakeBug(msg)

        if self.method == "pickle":
            try:
                return safe_pickle_load(filename)
            except Exception as e:
                msg = f"Could not unpickle data for key {key!r}. \n file: {filename}"
                logger.error(msg)
                logger.exception(e)
                msg += "\n" + traceback.format_exc()
                raise CompmakeBug(msg)
        elif self.method == "dill":
            with safe_read(filename) as f:
                return dill.load(f)
        else:
            raise NotImplementedError(self.method)

    def check_existence(self) -> None:
        if not self.checked_existence:
            self.checked_existence = True
            if not os.path.exists(self.basepath):
                # logger.info('Creating filesystem db %r' % self.basepath)
                os.makedirs(self.basepath)

    @track_time
    def __setitem__(self, key: StorageKey, value: object) -> None:  # @ReservedAssignment
        if trace_queries:
            logger.debug(f"W {str(key)}")

        self.check_existence()

        filename = self.filename_for_key(key)

        if self.method == "pickle":
            try:
                safe_pickle_dump(value, filename)
                assert os.path.exists(filename)
            except KeyboardInterrupt:
                raise
            except BaseException as e:
                msg = f"Cannot set key {key}: cannot pickle object of class {value.__class__.__name__}"
                logger.error(msg, e=traceback.format_exc())
                # logger.exception(e)
                emsg = find_pickling_error(value)
                logger.error(emsg)
                raise SerializationError(msg + "\n" + emsg)
        elif self.method == "dill":
            dill.settings["recurse"] = True
            dill.settings["byref"] = True
            with safe_write(filename) as f:
                return dill.dump(value, f)
        else:
            raise NotImplementedError(self.method)

    @track_time
    def __delitem__(self, key: StorageKey) -> None:
        filename = self.filename_for_key(key)
        if not os.path.exists(filename):
            msg = "I expected path %s to exist before deleting" % filename
            raise ValueError(msg)
        os.remove(filename)

    @track_time
    def __contains__(self, key: StorageKey) -> bool:
        if trace_queries:
            logger.debug(f"? {str(key)}")

        filename = self.filename_for_key(key)
        ex = os.path.exists(filename)

        # logger.debug('? %s %s %s' % (str(key), filename, ex))
        return ex

    @track_time
    def keys0(self, extension: str = None) -> Iterator[StorageKey]:
        if extension is None:
            extension = self.file_extension
        filename = self.filename_for_key("*", extension)
        for x in glob(filename):
            # b = splitext(basename(x))[0]
            b = basename(x.replace(extension, ""))
            key = self.basename2key(b)
            yield key

    @track_time
    def keys(self) -> List[StorageKey]:
        # slow process
        found = sorted(list(self.keys0()))
        return found

    def reopen_after_fork(self) -> None:
        pass

    dangerous_chars = {"/": "CMSLASH", "..": "CMDOT", "~": "CMHOME"}

    def key2basename(self, key):
        """ turns a key into a reasonable filename"""
        for char, replacement in self.dangerous_chars.items():
            key = key.replace(char, replacement)
        return key

    def basename2key(self, key):
        """ Undoes key2basename """
        for char, replacement in StorageFilesystem.dangerous_chars.items():
            key = key.replace(replacement, char)
        return key

    def filename_for_key(self, key, extension=None) -> FilePath:
        """ Returns the pickle storage filename corresponding to the job id """
        if extension is None:
            extension = self.file_extension
        f = self.key2basename(key) + extension
        return os.path.join(self.basepath, f)


def chmod_plus_x(filename: FilePath) -> None:
    st = os.stat(filename)
    os.chmod(filename, st.st_mode | stat.S_IEXEC)


def create_scripts(basepath: DirPath) -> None:
    filename2cmd = {
        "ls_failed": "ls failed",
        "why_failed": "why failed",
        "make_failed": "make failed",
        "remake": "remake",
        "why": "why",
        "make": "make",
        "parmake": "parmake",
        "rparmake": "rparmake",
        "rmake": "rmake",
        "clean": "clean",
        "ls": "ls",
        "stats": "stats",
        "gantt": "gantt",
        "details": "details",
    }
    for fn, cmd in filename2cmd.items():
        s = f'#!/bin/bash\ncompmake {basepath} -c "{cmd} $*"\n'
        f = os.path.join(basepath, fn)
        write_ustring_to_utf8_file(s, f, quiet=True)
        chmod_plus_x(f)

    s = f"#!/bin/bash\ncompmake {basepath} \n"
    f = os.path.join(basepath, "console")
    write_ustring_to_utf8_file(s, f, quiet=True)
    chmod_plus_x(f)

    s = f'#!/bin/bash\ncompmake {basepath} -c "$*" \n'
    f = os.path.join(basepath, "run")
    write_ustring_to_utf8_file(s, f, quiet=True)
    chmod_plus_x(f)



import sys

from .helpers import COMMANDS_ADVANCED, ui_command, UIState


__all__ = [
    "create_commands_html",
    "commands_html",
]


def create_commands_html(file=sys.stdout):  # @ReservedAssignment
    ordered_sections = sorted(UIState.sections.values(), key=lambda section_: section_.order)

    file.write("<table class='compmake-config'>\n")
    for section in ordered_sections:

        file.write("<tr><th colspan='3'>%s</th></tr/ \n" % section.name)

        if section.desc:
            file.write("<tr><td colspan='3'> %s </td></tr> \n" % section.desc)

        for name in section.commands:
            cmd = UIState.commands[name]
            short_doc = cmd.doc.split("\n")[0]

            # TODO
            params = ""

            file.write(
                "<tr> <td class='command-name'><tt>%s</tt></td> \
<td class='command-params'><tt>%s<tt></td> \
<td class='command-desc'>%s</td> </tr> \n"
                % (name, params, short_doc)
            )

    file.write("</table>\n")


@ui_command(section=COMMANDS_ADVANCED)
def commands_html(output_file=""):
    """ Dumps the commands description in html on the specified file. """
    if output_file:
        f = open(output_file, "w")
    else:
        f = sys.stdout
    create_commands_html(f)



# Warning: this is an auto-generated file
from .structures import EventSpec

compmake_registered_events = {}

__all__ = []


def add(e):
    compmake_registered_events[e.name] = e


add(EventSpec("compmake-init"))
add(EventSpec("compmake-closing"))

add(EventSpec("ui-message", ["string"]))
add(EventSpec("ui-status-summary", ["string"]))

add(EventSpec("job-stdout", ["job_id", "host", "lines"]))
add(EventSpec("job-stderr", ["job_id", "host", "lines"]))
add(EventSpec("job-progress", ["job_id", "host", "done", "progress", "goal"]))
add(EventSpec("job-progress-plus", ["job_id", "host", "stack"]))
add(EventSpec("job-succeeded", ["job_id", "host"]))
add(EventSpec("job-failed", ["job_id", "host", "reason", "bt"]))
# add(EventSpec("job-instanced", ["job_id", "host"]))
# add(EventSpec("job-starting", ["job_id", "host"]))
# add(EventSpec("job-finished", ["job_id", "host"]))
add(EventSpec("job-interrupted", ["job_id", "host", "bt"]))  # FIXME: nobody throws?
# add(EventSpec("job-now-ready", ["job_id"]))
add(EventSpec("manager-phase", ["phase"]))
add(
    EventSpec(
        "manager-loop",
        ["processing"],
        desc="called each time the manager loops waiting for jobs"
        "to finish. processing is the list of jobs currently "
        "processing.",
    )
)
# These are called when the manager updates its data structure
add(EventSpec("manager-job-processing", ["job_id"]))
add(EventSpec("manager-job-failed", ["job_id"]))
add(EventSpec("manager-job-blocked", ["job_id", "blocking_job_id"]))
add(EventSpec("manager-job-ready", ["job_id"]))
add(EventSpec("manager-job-done", ["job_id"]))
add(EventSpec("manager-host-failed", ["job_id", "host", "reason", "bt"]))
add(EventSpec("manager-init", ["targets", "more"]))
add(
    EventSpec(
        "manager-wait", ["reasons"], desc="Reasons why no jobs cannot be instantiated."  # dict str -> str
    )
)
add(
    EventSpec(
        "manager-progress",
        ["targets", "all_targets", "done", "todo", "failed", "ready", "processing", "deleted", "blocked"],
    )
)
add(
    EventSpec(
        "manager-succeeded",
        [
            "nothing_to_do",  # there was nothing to do (bool)
            "targets",
            "all_targets",
            "done",
            "todo",
            "failed",
            "ready",
            "processing",
            "blocked",
        ],
    )
)
add(
    EventSpec(
        "manager-interrupted",
        ["targets", "all_targets", "done", "todo", "failed", "ready", "processing", "blocked"],
    )
)
add(
    EventSpec(
        "manager-failed",
        ["reason", "targets", "all_targets", "done", "todo", "failed", "ready", "processing", "blocked"],
    )
)
add(EventSpec("worker-status", ["status", "job_id"]))
add(EventSpec("console-starting"))
add(EventSpec("console-ending"))

# These are called for commands strings ("make;clean")
add(EventSpec("command-line-starting", ["command"]))
add(EventSpec("command-line-failed", ["command", "retcode", "reason"]))
add(EventSpec("command-line-succeeded", ["command"]))
add(EventSpec("command-line-interrupted", ["command", "reason"]))


# These are called when a single command is executed
add(EventSpec("command-starting", ["command"]))
add(EventSpec("command-failed", ["command", "retcode", "reason"]))
add(EventSpec("command-succeeded", ["command"]))
add(EventSpec("command-interrupted", ["command", "reason", "traceback"]))

add(EventSpec("parmake-status", ["status"]))

add(EventSpec("job-defined", ["job_id"], desc="a new job is defined"))
add(EventSpec("job-already-defined", ["job_id"]))
add(EventSpec("job-redefined", ["job_id", "reason"]))

add(EventSpec("compmake-bug", ["user_msg", "dev_msg"]))



from compmake import Context, StorageFilesystem
from compmake.jobs.storage import all_jobs
from compmake.state import set_compmake_config
from compmake.ui.visualization import info
from nose.tools import istest
from tempfile import mkdtemp
import unittest


def g():
    pass


def h():
    pass


@istest
class Utils(unittest.TestCase):
    def all_jobs(self, root):
        """ Returns the list of jobs corresponding to the given expression. """
        db = StorageFilesystem(root, compress=True)
        return sorted(list(all_jobs(db)))


@istest
class TestCleaning1(Utils):
    def test_cleaning_other(self):
        root = mkdtemp()
        self.run_first(root)
        jobs1 = self.all_jobs(root)
        self.assertEqual(jobs1, ["g", "h"])
        self.run_second(root)
        jobs2 = self.all_jobs(root)
        self.assertEqual(jobs2, ["g"])

    def run_first(self, root):
        db = StorageFilesystem(root, compress=True)
        cc = Context(db=db)
        #
        cc.comp(g, job_id="g")
        cc.comp(h, job_id="h")
        cc.batch_command("make")

    def run_second(self, root):
        db = StorageFilesystem(root, compress=True)
        cc = Context(db=db)
        #
        cc.comp(g, job_id="g")
        cc.batch_command("make")


def f1(context):
    context.comp(g)
    context.comp(h)


def f2(context):
    context.comp(g)


@istest
class TestCleaning2(Utils):
    def test_cleaning2(self):
        root = mkdtemp()
        self.run_first(root)
        jobs1 = self.all_jobs(root)
        self.assertEqual(jobs1, ["f", "f-g", "f-h"])
        self.run_second(root)
        jobs2 = self.all_jobs(root)
        self.assertEqual(jobs2, ["f", "f-g"])

    def run_first(self, root):
        info("run_first()")
        db = StorageFilesystem(root, compress=True)
        cc = Context(db=db)
        #
        cc.comp_dynamic(f1, job_id="f")
        cc.batch_command("make recurse=1")

    def run_second(self, root):
        info("run_second()")
        db = StorageFilesystem(root, compress=True)
        cc = Context(db=db)
        #
        cc.comp_dynamic(f2, job_id="f")
        cc.batch_command("clean;make recurse=1")


def e1(context):
    context.comp_dynamic(f1, job_id="f")


def e2(context):
    context.comp_dynamic(f2, job_id="f")


@istest
class TestCleaning3(Utils):
    """ Now with multi level """

    #     @expected_failure
    def test_cleaning3(self):
        set_compmake_config("check_params", True)
        root = mkdtemp()
        self.run_first(root)
        jobs1 = self.all_jobs(root)
        self.assertEqual(jobs1, ["e", "f", "f-g", "f-h"])
        self.run_second(root)
        jobs2 = self.all_jobs(root)
        self.assertEqual(jobs2, ["e", "f", "f-g"])

    def run_first(self, root):
        print("run_first()")
        db = StorageFilesystem(root, compress=True)
        cc = Context(db=db)
        #
        cc.comp_dynamic(e1, job_id="e")
        cc.batch_command("make recurse=1; ls")

    def run_second(self, root):
        print("run_second()")
        db = StorageFilesystem(root, compress=True)
        cc = Context(db=db)
        #
        cc.comp_dynamic(e2, job_id="e")
        cc.batch_command("details e;clean;ls;make recurse=1")



from .. import CompmakeGlobalState, ConfigSection, ConfigSwitch, get_compmake_config, set_compmake_config
from ..exceptions import UserError

from ..utils import interpret_strings_like  # XXX initializtion order


def add_config_switch(name, default_value, allowed=None, desc=None, section=None, order=0):
    config_switches = CompmakeGlobalState.config_switches
    config_sections = CompmakeGlobalState.config_sections

    if name in config_switches:
        raise ValueError("Switch %r already defined" % name)

    config_switches[name] = ConfigSwitch(
        name=name, default_value=default_value, desc=desc, section=section, order=order, allowed=allowed
    )

    set_compmake_config(name, default_value)

    if not section in config_sections:
        raise ValueError("Section %r not defined." % section)

    config_sections[section].switches.append(name)


def set_config_from_strings(name, args):
    """ Sets config from an array of arguments """
    config_switches = CompmakeGlobalState.config_switches
    if not name in config_switches:
        raise UserError("I don't know config switch %r." % name)

    switch = config_switches[name]
    try:
        value = interpret_strings_like(args, switch.default_value)
    except ValueError as e:
        raise UserError(str(e)) from None

    set_compmake_config(name, value)


def add_config_section(name, desc=None, order=0):
    config_sections = CompmakeGlobalState.config_sections

    if name in config_sections:
        raise ValueError("Section %r already defined." % name)

    config_sections[name] = ConfigSection(name=name, desc=desc, order=order, switches=[])


def show_config() -> str:
    config_sections = CompmakeGlobalState.config_sections
    config_switches = CompmakeGlobalState.config_switches

    ordered_sections = sorted(config_sections.values(), key=lambda _section: _section.order)

    max_len_name = 1 + max([len(s.name) for s in config_switches.values()])
    max_len_val = 1 + max([len(str(get_compmake_config(s.name))) for s in config_switches.values()])

    b = ""
    for section in ordered_sections:
        b += "  ---- %s ----  \n" % section.name
        if section.desc:
            # XXX  multiline
            b += "  | %s \n" % section.desc
        for name in section.switches:
            switch = config_switches[name]
            value = get_compmake_config(name)
            changed = value != switch.default_value
            value = str(value)
            desc = str(switch.desc)

            if changed:
                attrs = ["bold"]
                if not get_compmake_config("colorize"):
                    value += "*"
            else:
                attrs = []

            from compmake.utils.colored import compmake_colored

            s1 = compmake_colored(name.rjust(max_len_name), attrs=["bold"])
            s2 = compmake_colored(value.rjust(max_len_val), attrs=attrs)
            b += "  | %s  %s  %s\n" % (s1, s2, desc)
    return b



from optparse import OptionGroup, OptionValueError

from .. import CompmakeGlobalState
from .structure import set_config_from_strings


__all__ = [
    "config_populate_optparser",
]


def config_populate_optparser(parser):
    config_switches = CompmakeGlobalState.config_switches
    config_sections = CompmakeGlobalState.config_sections

    for section in config_sections:  # section name -> ConfigSection
        config_section = config_sections[section]
        group = OptionGroup(parser, section, config_section.desc)

        switches = config_sections[section].switches

        for name in switches:
            switch = config_switches[name]

            command = "--%s" % switch.name

            group.add_option(
                command,
                nargs=1,
                help=switch.desc,
                type="string",
                action="callback",
                callback=option_callback,
                callback_kwargs={"switch": switch},
            )

        parser.add_option_group(group)


# noinspection PyUnusedLocal
def option_callback(option, opt, value, par, switch):
    try:
        set_config_from_strings(switch.name, value)
    except:
        raise OptionValueError('Could not parse value "%s" passed to "%s".' % (value, opt))



""" Contains queries of the job DB. """
import warnings
from contextlib import contextmanager
from typing import Collection, Set

from compmake.exceptions import CompmakeBug
from compmake.jobs.storage import get_job_cache
from compmake.structures import Cache, CMJobID
from zuper_commons.types import check_isinstance

from ..jobs import all_jobs, get_job

__all__ = [
    "parents",
    "direct_parents",
    "direct_children",
    "children",
    "top_targets",
    "tree",
    "jobs_defined",
    "definition_closure",
]


@contextmanager
def trace_bugs(msg):
    try:
        yield
    except CompmakeBug as e:
        raise CompmakeBug(msg) from e


# @contract(returns="set(unicode)")
def jobs_defined(job_id: CMJobID, db) -> Set[CMJobID]:
    """
        Gets the jobs defined by the given job.
        The job must be DONE.
    """
    check_isinstance(job_id, str)
    with trace_bugs("jobs_defined(%r)" % job_id):
        cache = get_job_cache(job_id, db=db)
        if cache.state != Cache.DONE:
            msg = "Cannot get jobs_defined for job not done " + "(status: %s)" % Cache.state2desc[cache.state]
            raise CompmakeBug(msg)
        return set(cache.jobs_defined)


from . import logger


# @contract(jobs="Iterable", returns="set(unicode)")
def definition_closure(jobs: Collection[CMJobID], db) -> Set[CMJobID]:
    """ The result does not contain jobs (unless one job defines another) """
    # print('definition_closure(%s)' % jobs)
    check_isinstance(jobs, (list, set))
    jobs = set(jobs)
    from compmake.jobs.uptodate import CacheQueryDB

    cq = CacheQueryDB(db)
    stack = set(jobs)
    result = set()
    while stack:
        # print('stack: %s' % stack)
        a = stack.pop()
        if not cq.job_exists(a):
            logger.warning("Warning: job %r does not exist anymore; ignoring." % a)
            continue

        if cq.get_job_cache(a).state == Cache.DONE:
            a_d = cq.jobs_defined(a)
            # print('%s ->%s' % (a, a_d))
            for x in a_d:
                result.add(x)
                stack.add(x)

    # print('  result = %s' % result)
    return result


def direct_parents(job_id: CMJobID, db) -> Set[CMJobID]:
    """ Returns the direct parents of the specified job.
        (Jobs that depend directly on this one) """
    check_isinstance(job_id, str)
    with trace_bugs(f"direct_parents({job_id!r})"):
        computation = get_job(job_id, db=db)
        return set(computation.parents)


def direct_children(job_id: CMJobID, db) -> Set[CMJobID]:
    """ Returns the direct children (dependencies) of the specified job """
    check_isinstance(job_id, str)
    with trace_bugs(f"direct_children({job_id!r})"):
        computation = get_job(job_id, db=db)
        return set(computation.children)


def children(job_id: CMJobID, db) -> Set[CMJobID]:
    """ Returns children, children of children, etc. """
    check_isinstance(job_id, str)
    with trace_bugs(f"children({job_id!r})"):
        t = set()
        for c in direct_children(job_id, db=db):
            t.add(c)
            t.update(children(c, db=db))
        return t


def top_targets(db):
    """ Returns a list of all jobs which are not needed by anybody """
    return [x for x in all_jobs(db=db) if not direct_parents(x, db=db)]


# def bottom_targets(db):
# """ Returns a list of all jobs with no dependencies. """
# return [x for x in all_jobs(db=db) if not direct_children(x, db=db)]


# @contract(jobs="list|set")
def tree(jobs: Collection[CMJobID], db):
    """
        Returns the tree of all dependencies of the jobs.
        Note this is very inefficient because recursive.
    """
    warnings.warn("Do not use -- very inefficient")
    t = set(jobs)
    for job_id in jobs:
        children_id = direct_children(job_id, db=db)
        t = t.union(tree(children_id, db=db))
    return t


def parents(job_id: CMJobID, db):
    """ Returns the set of all the parents, grandparents, etc.
        (does not include job_id) """
    check_isinstance(job_id, str)

    with trace_bugs(f"parents({job_id!r})"):
        t = set()
        parents_jobs = direct_parents(job_id, db=db)
        for p in parents_jobs:
            t.add(p)
            t.update(parents(p, db=db))
        return t
